# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides various dialog-based Output classes"""

from image_creator.output import Output
import time
import fcntl


class GaugeOutput(Output):
    """Output class implemented using dialog's gauge widget"""
    def __init__(self, dialog, title, msg=''):
        self.d = dialog
        self.msg = msg
        self.percent = 0
        self.d.gauge_start(self.msg, title=title)

        # Open pipe workaround. A fork will duplicate the open file descriptor.
        # The FD_CLOEXEC flag makes sure that the gauge internal fd will be
        # closed if execve is executed. This is needed because libguestfs will
        # fork/exec the KVM process. If this fd stays open in the KVM process,
        # then doing a gauge_stop will make this process wait forever for
        # a dialog process that is blocked waiting for input from the KVM
        # process's open file descriptor.
        fd = self.d._gauge_process['stdin'].fileno()
        flags = fcntl.fcntl(fd, fcntl.F_GETFL, 0)
        fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)

    def output(self, msg='', new_line=True):
        """Print msg as normal output"""
        self.msg = msg
        self.percent = 0
        self.d.gauge_update(self.percent, self.msg, update_text=True)
        time.sleep(0.4)

    def success(self, result, new_line=True):
        """Print result after a successful action"""
        self.percent = 100
        self.d.gauge_update(self.percent, "%s %s" % (self.msg, result),
                            update_text=True)
        time.sleep(0.4)

    def warn(self, msg, new_line=True):
        """Print a warning"""
        self.d.gauge_update(self.percent, "%s Warning: %s" % (self.msg, msg),
                            update_text=True)
        time.sleep(0.4)

    def cleanup(self):
        """Cleanup the GaugeOutput instance"""
        self.d.gauge_stop()

    class _Progress(Output._Progress):
        """Progress class for dialog's gauge widget"""
        template = {
            'default': '%(index)d/%(size)d',
            'percent': '',
            'b': "%(index)d/%(size)d B",
            'kb': "%(index)d/%(size)d KB",
            'mb': "%(index)d/%(size)d MB"
        }

        def __init__(self, size, title, bar_type='default'):
            self.output.size = size
            self.bar_type = bar_type
            self.output.msg = "%s ..." % title
            self.goto(0)

        def _postfix(self):
            return self.template[self.bar_type] % self.output.__dict__

        def goto(self, dest):
            """Move progress bar to a specific position"""
            self.output.index = dest
            self.output.percent = self.output.index * 100 // self.output.size
            msg = "%s %s" % (self.output.msg, self._postfix())
            self.output.d.gauge_update(self.output.percent, msg,
                                       update_text=True)

        def next(self):
            """Move progress bar one step forward"""
            self.goto(self.output.index + 1)


class InfoBoxOutput(Output):
    """Output class implemented using dialog's infobox widget"""
    def __init__(self, dialog, title, msg='', height=20, width=70):
        self.d = dialog
        self.title = title
        self.msg = msg
        self.width = width
        self.height = height
        self.d.infobox(self.msg, title=self.title)

    def output(self, msg='', new_line=True):
        """Print msg as normal output"""
        nl = '\n' if new_line else ''
        self.msg += "%s%s" % (msg, nl)
        # If output is long, only output the last lines that fit in the box
        lines = self.msg.splitlines()
        # The height of the active region is 2 lines shorter that the height of
        # the dialog
        h = self.height - 2
        display = self.msg if len(lines) <= h else "\n".join(lines[-h:])
        self.d.infobox(display, title=self.title, height=self.height,
                       width=self.width)

    def success(self, result, new_line=True):
        """Print result after an action is completed successfully"""
        self.output(result, new_line)

    def warn(self, msg, new_line=True):
        """Print a warning message"""
        self.output("Warning: %s" % msg, new_line)

    def finalize(self):
        """Finalize the output. After this is called, the InfoboxOutput
        instance should be destroyed
        """
        self.d.msgbox(self.msg, title=self.title, height=(self.height + 2),
                      width=self.width)

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides various helper functions to be used by other parts of
the package.
"""

import sh
import time
import os
import re
import json
import tempfile


class FatalError(Exception):
    """Fatal Error exception of snf-image-creator"""
    pass


def get_command(command):
    """Return a file system binary command"""
    def find_sbin_command(command, exception):
        search_paths = ['/usr/local/sbin', '/usr/sbin', '/sbin']
        for fullpath in map(lambda x: "%s/%s" % (x, command), search_paths):
            if os.path.exists(fullpath) and os.access(fullpath, os.X_OK):
                return sh.Command(fullpath)
        raise exception

    try:
        return sh.__getattr__(command)
    except sh.CommandNotFound as e:
        return find_sbin_command(command, e)


def image_info(image):
    """Returns information about an image file"""

    qemu_img = get_command('qemu-img')
    try:
        info = qemu_img('info', '--output', 'json', image)
    except sh.ErrorReturnCode_1:
        # Old version of qemu-img that does not support --output json
        info = qemu_img('info', image)
        for line in str(info).splitlines():
            if line.startswith('file format:'):
                format = line.split(':')[1].strip()
                return {'format': format}
        raise FatalError("Unable to determine the image format")
    return json.loads(str(info))


def create_snapshot(source, target_dir):
    """Returns a qcow2 snapshot of an image file"""

    qemu_img = get_command('qemu-img')
    snapfd, snap = tempfile.mkstemp(prefix='snapshot-', dir=target_dir)
    os.close(snapfd)
    qemu_img('create', '-f', 'qcow2', '-o',
             'backing_file=%s' % os.path.abspath(source), snap)
    return snap


def get_kvm_binary():
    """Returns the path to the kvm binary and some extra arguments if needed"""

    uname = get_command('uname')
    which = get_command('which')

    machine = str(uname('-m')).strip()
    if re.match('i[3-6]86', machine):
        machine = 'i386'

    binary = which('qemu-system-%s' % machine)

    needed_args = "--enable-kvm",

    if binary is None:
        return which('kvm'), tuple()

    return binary, needed_args


def try_fail_repeat(command, *args):
    """Execute a command multiple times until it succeeds"""
    times = (0.1, 0.5, 1, 2)
    i = iter(times)
    while True:
        try:
            command(*args)
            return
        except sh.ErrorReturnCode:
            try:
                wait = i.next()
            except StopIteration:
                break
            time.sleep(wait)

    raise FatalError("Command: `%s %s' failed" % (command, " ".join(args)))


def free_space(dirname):
    """Compute the free space in a directory"""
    stat = os.statvfs(dirname)
    return stat.f_bavail * stat.f_frsize


def virtio_versions(virtio_state):
    """Returns the versions of the drivers defined by the virtio state"""

    ret = {}
    for name, infs in virtio_state.items():
        driver_ver = [drv['DriverVer'].split(',', 1) if 'DriverVer' in drv
                      else [] for drv in infs.values()]
        vers = [v[1] if len(v) > 1 else " " for v in driver_ver]
        ret[name] = "<not found>" if len(infs) == 0 else ", ".join(vers)

    return ret


class QemuNBD(object):
    """Wrapper class for the qemu-nbd tool"""

    def __init__(self, image):
        """Initialize an instance"""
        self.image = image
        self.device = None
        self.pattern = re.compile('^nbd\d+$')
        self.modprobe = get_command('modprobe')
        self.qemu_nbd = get_command('qemu-nbd')

    def _list_devices(self):
        """Returns all the NBD block devices"""
        return set([d for d in os.listdir('/dev/') if self.pattern.match(d)])

    def connect(self, ro=True):
        """Connect the image to a free NBD device"""
        devs = self._list_devices()

        if len(devs) == 0:  # Is nbd module loaded?
            self.modprobe('nbd', 'max_part=16')
            # Wait a second for /dev to be populated
            time.sleep(1)
            devs = self._list_devices()
            if len(devs) == 0:
                raise FatalError("/dev/nbd* devices not present!")

        # Ignore the nbd block devices that are in use
        with open('/proc/partitions') as partitions:
            for line in iter(partitions):
                entry = line.split()
                if len(entry) != 4:
                    continue
                if entry[3] in devs:
                    devs.remove(entry[3])

        if len(devs) == 0:
            raise FatalError("All NBD block devices are busy!")

        device = '/dev/%s' % devs.pop()
        args = ['-c', device]
        if ro:
            args.append('-r')
        args.append(self.image)

        self.qemu_nbd(*args)
        self.device = device
        return device

    def disconnect(self):
        """Disconnect the image from the connected device"""
        assert self.device is not None, "No device connected"

        self.qemu_nbd('-d', self.device)
        self.device = None

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module hosts OS-specific code common to all Unix-like OSs."""

from image_creator.os_type import OSBase, sysprep


class Unix(OSBase):
    """OS class for Unix"""
    sensitive_userdata = [
        '.history',
        '.sh_history',
        '.bash_history',
        '.zsh_history',
        '.gnupg',
        '.ssh',
        '.kamakirc',
        '.kamaki.history',
        '.kamaki.log'
    ]

    def _mountpoints(self):
        """Return mountpoints in the correct order.
        / should be mounted before /boot or /usr, /usr befor /usr/bin ...
        """
        mps = self.image.g.inspect_get_mountpoints(self.root)

        def compare(a, b):
            if len(a[0]) > len(b[0]):
                return 1
            elif len(a[0]) == len(b[0]):
                return 0
            else:
                return -1
        mps.sort(compare)

        for mp in mps:
            yield mp

    def _do_mount(self, readonly):
        """Mount partitions in the correct order"""

        critical_mpoints = ('/', '/etc', '/root', '/home', '/var')

        mopts = 'ro' if readonly else 'rw'
        for mp, dev in self._mountpoints():
            try:
                self.image.g.mount_options(mopts, dev, mp)
            except RuntimeError as msg:
                if mp in critical_mpoints:
                    self._mount_error = str(msg)
                    return False
                else:
                    self._mount_warnings.append('%s (ignored)' % msg)

        return True

    @sysprep('Removing files under /var/cache')
    def _cleanup_cache(self):
        """Remove all regular files under /var/cache"""

        self._foreach_file('/var/cache', self.image.g.rm, ftype='r')

    @sysprep('Removing files under /tmp and /var/tmp')
    def _cleanup_tmp(self):
        """Remove all files under /tmp and /var/tmp"""

        self._foreach_file('/tmp', self.image.g.rm_rf, maxdepth=1)
        self._foreach_file('/var/tmp', self.image.g.rm_rf, maxdepth=1)

    @sysprep('Emptying all files under /var/log')
    def _cleanup_log(self):
        """Empty all files under /var/log"""

        self._foreach_file('/var/log', self.image.g.truncate, ftype='r')

    @sysprep('Removing files under /var/mail & /var/spool/mail', enabled=False)
    def _cleanup_mail(self):
        """Remove all files under /var/mail and /var/spool/mail"""

        self._foreach_file('/var/spool/mail', self.image.g.rm_rf, maxdepth=1)

        self._foreach_file('/var/mail', self.image.g.rm_rf, maxdepth=1)

    @sysprep('Removing sensitive user data')
    def _cleanup_userdata(self):
        """Delete sensitive user data"""

        homedirs = ['/root']
        if self.image.g.is_dir('/home/'):
            homedirs += self._ls('/home/')

        action = self.image.g.rm_rf
        if self._scrub_support:
            action = self.image.g.scrub_file
        else:
            self.out.warn("Sensitive data won't be scrubbed (not supported)")

        for homedir in homedirs:
            for data in self.sensitive_userdata:
                fname = "%s/%s" % (homedir, data)
                if self.image.g.is_file(fname):
                    action(fname)
                elif self.image.g.is_dir(fname):
                    self._foreach_file(fname, action, ftype='r')

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides an interface for launching windows VMs"""

import random
import subprocess
import signal
import tempfile
import os
import time
import errno
from string import lowercase, uppercase, digits

from image_creator.util import FatalError, get_kvm_binary
from image_creator.os_type.windows.winexe import WinEXE, WinexeTimeout

# Just a random 16 character long token
RANDOM_TOKEN = "".join(random.choice(lowercase + uppercase) for _ in range(16))


class VM(object):
    """Windows Virtual Machine"""
    def __init__(self, disk, params, admin):
        """Create VM instance"""

        self.disk = disk
        self.params = params
        self.admin = admin
        self.interface = 'virtio'

        # expected number of token occurrences in serial port
        self._ntokens = 0

        kvm, needed_args = get_kvm_binary()
        if kvm is None:
            raise FatalError("Can't find the kvm binary")

        self.kvm = [kvm] + list(needed_args)

        def random_mac():
            """creates a random mac address"""
            mac = [0x00, 0x16, 0x3e,
                   random.randint(0x00, 0x7f),
                   random.randint(0x00, 0xff),
                   random.randint(0x00, 0xff)]

            return ':'.join(['%02x' % x for x in mac])

        self.mac = random_mac()

        def random_password():
            """Creates a random password"""

            # I borrowed this from Synnefo
            pool = lowercase + uppercase + digits
            lowerset = set(lowercase)
            upperset = set(uppercase)
            digitset = set(digits)
            length = 10

            password = ''.join(random.choice(pool) for i in range(length - 2))

            # Make sure the password is compliant
            chars = set(password)
            if not chars & lowerset:
                password += random.choice(lowercase)
            if not chars & upperset:
                password += random.choice(uppercase)
            if not chars & digitset:
                password += random.choice(digits)

            # Pad if necessary to reach required length
            password += ''.join(random.choice(pool) for i in
                                range(length - len(password)))

            return password

        self.password = random_password()

        # Use Ganeti's VNC port range for a random vnc port
        self.display = random.randint(11000, 14999) - 5900

        self.serial = None
        self.process = None

    def isalive(self):
        """Check if the VM is alive"""
        return self.process is not None and self.process.poll() is None

    def start(self, **kwargs):
        """Start the windows VM"""

        self._ntokens = 0

        args = []
        args.extend(self.kvm)

        if 'smp' in self.params:
            args.extend(['-smp', str(self.params['smp'].value)])

        if 'mem' in self.params:
            args.extend(['-m', str(self.params['mem'].value)])

        args.extend(['-drive',
                     'file=%s,format=raw,cache=unsafe,if=%s' %
                     (self.disk, self.interface)])

        args.extend(
            ['-netdev', 'type=user,hostfwd=tcp::445-:445,id=netdev0',
             '-device', 'rtl8139,mac=%s,netdev=netdev0' % self.mac])

        if 'extra_disk' in kwargs:
            fname, iftype = kwargs['extra_disk']
            args.extend(['-drive',
                         'file=%s,format=raw,cache=unsafe,if=%s' %
                         (fname, iftype)])

        args.extend(['-vnc', ":%d" % self.display])

        # The serial port
        serialfd, self.serial = tempfile.mkstemp()
        os.close(serialfd)
        args.extend(['-serial', 'file:%s' % self.serial])

        args.extend(['-monitor', 'stdio'])

        self.process = subprocess.Popen(args, stdin=subprocess.PIPE,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE)

    def stop(self, timeout=0, fatal=True):
        """Stop the VM"""

        try:
            if not self.isalive():
                return

            def handler(signum, frame):
                self.process.terminate()
                time.sleep(1)
                if self.isalive():
                    self.process.kill()
                self.process.wait()
                if fatal:
                    raise FatalError("Stopping the VM timed-out")

            signal.signal(signal.SIGALRM, handler)

            signal.alarm(timeout)
            self.process.communicate(input="system_powerdown\n")
            signal.alarm(0)

        finally:
            if self.serial is not None:
                try:
                    os.unlink(self.serial)
                except OSError as e:
                    if errno.errorcode[e.errno] != 'ENOENT':  # File not found
                        raise

    def wait_on_serial(self, timeout):
        """Wait until the random token appears on the VM's serial port"""

        self._ntokens += 1

        for _ in xrange(timeout):
            time.sleep(1)
            with open(self.serial) as f:
                current = 0
                for line in f:
                    if line.startswith(RANDOM_TOKEN):
                        current += 1
                        if current == self._ntokens:
                            return True
            if not self.isalive():
                (stdout, stderr, rc) = self.wait()
                raise FatalError("Windows VM died unexpectedly!\n\n"
                                 "(rc=%d)\n%s" % (rc, stderr))

        return False

    def wait(self, timeout=0):
        """Wait for the VM to shutdown by itself"""

        def handler(signum, frame):
            raise FatalError("VM wait timed-out.")

        signal.signal(signal.SIGALRM, handler)

        signal.alarm(timeout)
        stdout, stderr = self.process.communicate()
        signal.alarm(0)

        return (stdout, stderr, self.process.poll())

    def rexec(self, command, **kwargs):
        """Remote execute a command on the windows VM

        The following optional flags are allowed:

        * fatal: If True, a FatalError is thrown if the command fails

        * debug: If True, WinEXE is executed in the highest debug level

        * uninstall: If True, the winexesvc.exe service will be uninstalled
          after the execution of the command.
        """

        fatal = kwargs['fatal'] if 'fatal' in kwargs else True
        debug = kwargs['debug'] if 'debug' in kwargs else False
        uninstall = kwargs['uninstall'] if 'uninstall' in kwargs else False

        winexe = WinEXE(self.admin.name, 'localhost', password=self.password)
        winexe.no_pass()

        if debug:
            winexe.debug(9)

        if uninstall:
            winexe.uninstall()

        try:
            (stdout, stderr, rc) = winexe.run(command)
        except WinexeTimeout:
            raise FatalError("Command: `%s' timeout out." % command)

        if rc != 0 and fatal:
            log = tempfile.NamedTemporaryFile(delete=False)
            try:
                log.file.write("STDOUT:\n%s\n" % stdout)
                log.file.write("STDERR:\n%s\n" % stderr)
            finally:
                fname = log.name
                log.close()

            # self.out.output("Command: `%s' failed (rc=%d). Reason: %s" %
            #                 (command, rc, reason))
            raise FatalError("Command: `%s' failed (rc=%d). See: %s" %
                             (command, rc, fname))

        return (stdout, stderr, rc)

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module providing useful functions for the dialog-based version of
snf-image-creator.
"""

import os
import stat
import re
import json
from image_creator.output.dialog import GaugeOutput
from image_creator.kamaki_wrapper import Kamaki

SMALL_WIDTH = 60
WIDTH = 70


def select_file(d, **kwargs):
    """Select a file or directory.

    The following optional arguments can be applied:

    * init: Initial file path. If this path is valid this will be returned

    * ftype: Allowed file types. If the value of this argument is "br" only
             block devices and regular files are valid. For a list of available
             file types, see here:
             http://libguestfs.org/guestfs.3.html#guestfs_readdir

    * title: The dialog box title. The default one is: "Please select a file"

    * bundle_host: This can be True or False. If this is True, an extra
      "Bundle Host" button will be present if the file selection dialog.
    """

    type_check = {'b': stat.S_ISBLK,   # Block special
                  'c': stat.S_ISCHR,   # Char special
                  'd': stat.S_ISDIR,   # Directory
                  'f': stat.S_ISFIFO,  # FIFO (named pipe)
                  'l': stat.S_ISLNK,   # Symbolic link
                  'r': stat.S_ISREG,   # Regular file
                  's': stat.S_ISSOCK}  # Socket

    fname = None if "init" not in kwargs else kwargs['init']
    ftype = set(t for t in kwargs['ftype']) if 'ftype' in kwargs else set('r')
    title = kwargs['title'] if 'title' in kwargs else 'Please select a file.'

    bundle_host = kwargs['bundle_host'] if 'bundle_host' in kwargs else None
    extra_button = 1 if bundle_host else 0

    for t in ftype:
        assert t in type_check, "Invalid ftype: %s" % t

    # This is a special case
    if bundle_host and fname == os.sep:
        return os.sep

    default = os.getcwd() + os.sep

    while 1:
        if fname is not None:
            if not os.path.exists(fname):
                d.msgbox("The file `%s' you choose does not exist." % fname,
                         width=SMALL_WIDTH)
            else:
                mode = os.stat(fname).st_mode
                for i in ftype:
                    if type_check[i](mode):
                        return fname

                if stat.S_ISDIR(mode):
                    default = fname
                else:
                    d.msgbox("Invalid input.", width=SMALL_WIDTH)

        (code, fname) = d.fselect(default, 10, 60, extra_button=extra_button,
                                  title=title, extra_label="Bundle Host")
        if code in (d.DIALOG_CANCEL, d.DIALOG_ESC):
            return None
        elif code == d.DIALOG_EXTRA:
            return os.sep

    return fname


def update_background_title(session):
    """Update the background title of the dialog page"""
    d = session['dialog']
    disk = session['disk']
    image = session['image']

    MB = 2 ** 20

    size = (image.size + MB - 1) // MB
    postfix = " (shrinked)" if image.os.shrinked else ''

    title = "OS: %s, Distro: %s, Size: %dMB%s, Source: %s" % \
            (image.ostype.capitalize(), image.distro.capitalize(), size,
             postfix, os.path.abspath(disk.source))

    d.setBackgroundTitle(title)


def confirm_exit(d, msg=''):
    """Ask the user to confirm when exiting the program"""
    return not d.yesno("%s Do you want to exit?" % msg, width=SMALL_WIDTH)


def confirm_reset(d):
    """Ask the user to confirm a reset action"""
    return not d.yesno("Are you sure you want to reset everything?",
                       width=SMALL_WIDTH, defaultno=1)


class Reset(Exception):
    """Exception used to reset the program"""
    pass


def extract_metadata_string(session):
    """Convert image metadata to text"""
    metadata = {}
    metadata.update(session['metadata'])
    if 'task_metadata' in session:
        for key in session['task_metadata']:
            metadata[key] = 'yes'

    return unicode(json.dumps({'properties': metadata,
                               'disk-format': 'diskdump'}, ensure_ascii=False))


def extract_image(session):
    """Dump the image to a local file"""
    d = session['dialog']
    dir = os.getcwd()
    while 1:
        if dir and dir[-1] != os.sep:
            dir = dir + os.sep

        (code, path) = d.fselect(dir, 10, 50, title="Save image as...")
        if code in (d.DIALOG_CANCEL, d.DIALOG_ESC):
            return False

        if os.path.isdir(path):
            dir = path
            continue

        if os.path.isdir("%s.meta" % path):
            d.msgbox("Can't overwrite directory `%s.meta'" % path,
                     width=SMALL_WIDTH)
            continue

        if os.path.isdir("%s.md5sum" % path):
            d.msgbox("Can't overwrite directory `%s.md5sum'" % path,
                     width=SMALL_WIDTH)
            continue

        basedir = os.path.dirname(path)
        name = os.path.basename(path)
        if not os.path.exists(basedir):
            d.msgbox("Directory `%s' does not exist" % basedir,
                     width=SMALL_WIDTH)
            continue

        dir = basedir
        if len(name) == 0:
            continue

        files = ["%s%s" % (path, ext) for ext in ('', '.meta', '.md5sum')]
        overwrite = filter(os.path.exists, files)

        if len(overwrite) > 0:
            if d.yesno("The following file(s) exist:\n"
                       "%s\nDo you want to overwrite them?" %
                       "\n".join(overwrite), width=SMALL_WIDTH):
                continue

        gauge = GaugeOutput(d, "Image Extraction", "Extracting image...")
        try:
            image = session['image']
            out = image.out
            out.add(gauge)
            try:
                if "checksum" not in session:
                    session['checksum'] = image.md5()

                # Extract image file
                image.dump(path)

                # Extract metadata file
                out.output("Extracting metadata file ...")
                with open('%s.meta' % path, 'w') as f:
                    f.write(extract_metadata_string(session))
                out.success('done')

                # Extract md5sum file
                out.output("Extracting md5sum file ...")
                md5str = "%s %s\n" % (session['checksum'], name)
                with open('%s.md5sum' % path, 'w') as f:
                    f.write(md5str)
                out.success("done")
            finally:
                out.remove(gauge)
        finally:
            gauge.cleanup()
        d.msgbox("Image file `%s' was successfully extracted!" % path,
                 width=SMALL_WIDTH)
        break

    return True


def _check_cloud(session, name, url, token):
    """Checks if the provided info for a cloud are valid"""
    d = session['dialog']
    regexp = re.compile(r'^[~@#$:\-\w]+$')

    if not re.match(regexp, name):
        d.msgbox("Allowed characters for name: a-zA-Z0-9_~@#$:-", width=WIDTH)
        return False

    if len(url) == 0:
        d.msgbox("URL cannot be empty!", width=WIDTH)
        return False

    if len(token) == 0:
        d.msgbox("Token cannot be empty!", width=WIDTH)
        return False

    if Kamaki.create_account(url, token) is None:
        d.msgbox("The cloud info you provided is not valid. Please check the "
                 "Authentication URL and the token values again!", width=WIDTH)
        return False

    return True


def add_cloud(session):
    """Add a new cloud account"""

    d = session['dialog']

    name = ""
    description = ""
    url = ""
    token = ""

    while 1:
        fields = [
            ("Name:", name, 60),
            ("Description (optional): ", description, 80),
            ("Authentication URL: ", url, 200),
            ("Token:", token, 100)]

        (code, output) = d.form("Add a new cloud account:", height=13,
                                width=WIDTH, form_height=4, fields=fields)

        if code in (d.DIALOG_CANCEL, d.DIALOG_ESC):
            return False

        name, description, url, token = output

        name = name.strip()
        description = description.strip()
        url = url.strip()
        token = token.strip()

        if _check_cloud(session, name, url, token):
            if name in Kamaki.get_clouds().keys():
                d.msgbox("A cloud with name `%s' already exists. If you want "
                         "to edit the existing cloud account, use the edit "
                         "menu." % name, width=WIDTH)
            else:
                Kamaki.save_cloud(name, url, token, description)
                break

        continue

    return True


def edit_cloud(session, name):
    """Edit a cloud account"""

    info = Kamaki.get_cloud_by_name(name)

    assert info, "Cloud: `%s' does not exist" % name

    description = info['description'] if 'description' in info else ""
    url = info['url'] if 'url' in info else ""
    token = info['token'] if 'token' in info else ""

    d = session['dialog']

    while 1:
        fields = [
            ("Description (optional): ", description, 80),
            ("Authentication URL: ", url, 200),
            ("Token:", token, 100)]

        (code, output) = d.form("Edit cloud account: `%s'" % name, height=13,
                                width=WIDTH, form_height=3, fields=fields)

        if code in (d.DIALOG_CANCEL, d.DIALOG_ESC):
            return False

        description, url, token = output

        description = description.strip()
        url = url.strip()
        token = token.strip()

        if _check_cloud(session, name, url, token):
            Kamaki.save_cloud(name, url, token, description)
            break

        continue

    return True


def update_sysprep_param(session, name, title=None):
    """Modify the value of a sysprep parameter"""
    d = session['dialog']
    image = session['image']

    param = image.os.sysprep_params[name]

    while 1:
        if param.type in ("file", "dir"):
            if not title:
                title = "Please select a %s to use for the `%s' parameter" % \
                    ('file' if param.type == 'file' else 'directory', name)
            ftype = "br" if param.type == 'file' else 'd'

            value = select_file(d, ftype=ftype, title=title)
            if value is None:
                return False
        else:
            if not title:
                title = "Please provide a new value for configuration " \
                        "parameter: `%s'" % name
            (code, answer) = d.inputbox(
                title, width=WIDTH, init=str(param.value))

            if code in (d.DIALOG_CANCEL, d.DIALOG_ESC):
                return False

            value = answer.strip()

        if param.set_value(value) is False:
            d.msgbox("Error: %s" % param.error, width=WIDTH)
            param.error = None
            continue
        break

    return True

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module hosts code to handle unknown OSs."""

from image_creator.os_type import OSBase


class Unsupported(OSBase):
    """OS class for unsupported OSs"""
    def __init__(self, image, **kargs):
        super(Unsupported, self).__init__(image, **kargs)

    def collect_metadata(self):
        """Collect metadata about the OS"""
        self.out.warn("Unable to collect metadata for unsupported media")

    def _do_mount(self, readonly):
        """Mount partitions in correct order"""
        self._mount_error = "not supported on this media"
        return False

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This package provides various classes for preparing different Operating
Systems for image creation.
"""

from image_creator.util import FatalError

import textwrap
import re
from collections import namedtuple
from functools import wraps


def os_cls(distro, osfamily):
    """Given the distro name and the osfamily, return the appropriate OSBase
    derived class
    """

    # hyphens are not allowed in module names
    canonicalize = lambda x: x.replace('-', '_').lower()

    distro = canonicalize(distro)
    osfamily = canonicalize(osfamily)

    try:
        module = __import__("image_creator.os_type.%s" % distro,
                            fromlist=['image_creator.os_type'])
        classname = distro.capitalize()
    except ImportError:
        try:
            module = __import__("image_creator.os_type.%s" % osfamily,
                                fromlist=['image_creator.os_type'])
            classname = osfamily.capitalize()
        except ImportError:
            raise FatalError("Unknown OS name: `%s'" % osfamily)

    return getattr(module, classname)


def add_prefix(target):
    """Decorator that adds a prefix to the result of a function"""
    def wrapper(self, *args):
        prefix = args[0]
        return [prefix + path for path in target(self, *args)]
    return wrapper


def sysprep(message, enabled=True, **kwargs):
    """Decorator for system preparation tasks"""
    def wrapper(method):
        assert method.__name__.startswith('_'), \
            "Invalid sysprep name:` %s'. Should start with _" % method.__name__

        method._sysprep = True
        method._sysprep_enabled = enabled
        method._sysprep_nomount = False

        for key, val in kwargs.items():
            setattr(method, "_sysprep_%s" % key, val)

        @wraps(method)
        def inner(self, print_message=True):
            if print_message:
                self.out.output(message)
            return method(self)

        return inner
    return wrapper


class SysprepParam(object):
    """This class represents a system preparation parameter"""

    def __init__(self, type, default, description, **kargs):

        assert hasattr(self, "_check_%s" % type), "Invalid type: %s" % type

        self.type = type
        self.default = default
        self.description = description
        self.value = default
        self.error = None
        self.check = kargs['check'] if 'check' in kargs else lambda x: x
        self.hidden = kargs['hidden'] if 'hidden' in kargs else False

    def set_value(self, value):
        """Update the value of the parameter"""

        check_type = getattr(self, "_check_%s" % self.type)
        try:
            self.value = self.check(check_type(value))
        except ValueError as e:
            self.error = e.message
            return False
        return True

    def _check_posint(self, value):
        """Check if the value is a positive integer"""
        try:
            value = int(value)
        except ValueError:
            raise ValueError("Invalid number")

        if value <= 0:
            raise ValueError("Value is negative or zero")

        return value

    def _check_string(self, value):
        """Check if a value is a string"""
        return str(value)

    def _check_file(self, value):
        """Check if the value is a valid filename"""

        value = str(value)
        if len(value) == 0:
            return ""

        import os

        def isblockdev(filename):
            import stat
            try:
                return stat.S_ISBLK(os.stat(filename).st_mode)
            except OSError:
                return False
        if os.path.isfile(value) or isblockdev(value):
            return value

        raise ValueError("Invalid filename")

    def _check_dir(self, value):
        """Check if the value is a valid directory"""

        value = str(value)
        if len(value) == 0:
            return ""

        import os
        if os.path.isdir(value):
            return value

        raise ValueError("Invalid dirname")


def add_sysprep_param(name, type, default, descr, **kargs):
    """Decorator for __init__ that adds the definition for a system preparation
    parameter in an instance of an os_type class
    """
    extra = kargs

    def wrapper(init):
        @wraps(init)
        def inner(self, *args, **kwargs):

            if not hasattr(self, 'sysprep_params'):
                self.sysprep_params = {}

            self.sysprep_params[name] = \
                SysprepParam(type, default, descr, **extra)
            init(self, *args, **kwargs)
        return inner
    return wrapper


def del_sysprep_param(name):
    """Decorator for __init__ that deletes a previously added sysprep parameter
    definition from an instance of a os_type class.
    """
    def wrapper(func):
        @wraps(func)
        def inner(self, *args, **kwargs):
            del self.sysprep_params[name]
            func(self, *args, **kwargs)
        return inner
    return wrapper


class OSBase(object):
    """Basic operating system class"""

    def __init__(self, image, **kargs):
        self.image = image

        self.root = image.root
        self.out = image.out

        # Could be defined in a decorator
        if not hasattr(self, 'sysprep_params'):
            self.sysprep_params = {}

        if 'sysprep_params' in kargs:
            for key, val in kargs['sysprep_params'].items():
                if key not in self.sysprep_params:
                    self.out.warn("Ignoring invalid `%s' parameter." % key)
                    continue
                param = self.sysprep_params[key]
                if not param.set_value(val):
                    raise FatalError("Invalid value for sysprep parameter: "
                                     "`%s'. Reason: %s" % (key, param.error))

        self.meta = {}
        self.shrinked = False

        # This will host the error if mount fails
        self._mount_error = ""
        self._mount_warnings = []
        self._mounted = False

        # Many guestfs compilations don't support scrub
        self._scrub_support = True
        try:
            self.image.g.available(['scrub'])
        except RuntimeError:
            self._scrub_support = False

        # Create a list of available syspreps
        self._sysprep_tasks = {}
        for name in dir(self):
            obj = getattr(self, name)
            if not hasattr(obj, '_sysprep'):
                continue
            self._sysprep_tasks[name] = obj._sysprep_enabled

        self._cleanup_jobs = {}

    def _add_cleanup(self, namespace, job, *args):
        """Add a new job in a cleanup list"""

        if namespace not in self._cleanup_jobs:
            self._cleanup_jobs[namespace] = []

        self._cleanup_jobs[namespace].append((job, args))

    def _cleanup(self, namespace):
        """Run the cleanup tasks that are defined under a specific namespace"""

        if namespace not in self._cleanup_jobs:
            self.out.warn("Cleanup namespace: `%s' is not defined", namespace)
            return

        while len(self._cleanup_jobs[namespace]):
            job, args = self._cleanup_jobs[namespace].pop()
            job(*args)

        del self._cleanup_jobs[namespace]

    def inspect(self):
        """Inspect the media to check if it is supported"""

        if self.image.is_unsupported():
            return

        self.out.output('Running OS inspection:')
        with self.mount(readonly=True, silent=True):
            self._do_inspect()
        self.out.output()

    def collect_metadata(self):
        """Collect metadata about the OS"""

        self.out.output('Collecting image metadata ...', False)

        with self.mount(readonly=True, silent=True):
            self._do_collect_metadata()

        self.out.success('done')
        self.out.output()

    def list_syspreps(self):
        """Returns a list of sysprep objects"""
        return [getattr(self, name) for name in self._sysprep_tasks]

    def sysprep_info(self, obj):
        """Returns information about a sysprep object"""
        assert hasattr(obj, '_sysprep'), "Object is not a sysprep"

        SysprepInfo = namedtuple("SysprepInfo", "name description display")

        name = obj.__name__.replace('_', '-')[1:]
        description = textwrap.dedent(obj.__doc__)
        display = getattr(obj, '_sysprep_display',
                          name.replace('-', ' ').capitalize())

        return SysprepInfo(name, description, display)

    def get_sysprep_by_name(self, name):
        """Returns the sysprep object with the given name"""
        error_msg = "Syprep operation %s does not exist for %s" % \
                    (name, self.__class__.__name__)

        method_name = '_' + name.replace('-', '_')

        if hasattr(self, method_name):
            method = getattr(self, method_name)

            if hasattr(method, '_sysprep'):
                return method

        raise FatalError(error_msg)

    def enable_sysprep(self, obj):
        """Enable a system preparation operation"""
        assert hasattr(obj, '_sysprep'), "Object is not a sysprep"
        assert obj.__name__ in self._sysprep_tasks, "Sysprep already executed"

        self._sysprep_tasks[obj.__name__] = True

    def disable_sysprep(self, obj):
        """Disable a system preparation operation"""
        assert hasattr(obj, '_sysprep'), "Object is not a sysprep"
        assert obj.__name__ in self._sysprep_tasks, "Sysprep already executed"

        self._sysprep_tasks[obj.__name__] = False

    def sysprep_enabled(self, obj):
        """Returns True if this system praparation operation is enabled"""
        assert hasattr(obj, '_sysprep'), "Object is not a sysprep"
        assert obj.__name__ in self._sysprep_tasks, "Sysprep already executed"

        return self._sysprep_tasks[obj.__name__]

    def print_syspreps(self):
        """Print enabled and disabled system preparation operations."""

        syspreps = self.list_syspreps()
        enabled = [s for s in syspreps if self.sysprep_enabled(s)]
        disabled = [s for s in syspreps if not self.sysprep_enabled(s)]

        wrapper = textwrap.TextWrapper()
        wrapper.subsequent_indent = '\t'
        wrapper.initial_indent = '\t'
        wrapper.width = 72

        self.out.output("Enabled system preparation operations:")
        if len(enabled) == 0:
            self.out.output("(none)")
        else:
            for sysprep in enabled:
                name = sysprep.__name__.replace('_', '-')[1:]
                descr = wrapper.fill(textwrap.dedent(sysprep.__doc__))
                self.out.output('    %s:\n%s\n' % (name, descr))

        self.out.output("Disabled system preparation operations:")
        if len(disabled) == 0:
            self.out.output("(none)")
        else:
            for sysprep in disabled:
                name = sysprep.__name__.replace('_', '-')[1:]
                descr = wrapper.fill(textwrap.dedent(sysprep.__doc__))
                self.out.output('    %s:\n%s\n' % (name, descr))

    def print_sysprep_params(self):
        """Print the system preparation parameter the user may use"""

        self.out.output("System preparation parameters:")
        self.out.output()

        public_params = [(n, p) for n, p in self.sysprep_params.items()
                         if not p.hidden]
        if len(public_params) == 0:
            self.out.output("(none)")
            return

        wrapper = textwrap.TextWrapper()
        wrapper.subsequent_indent = "             "
        wrapper.width = 72

        for name, param in public_params:
            if param.hidden:
                continue
            self.out.output("NAME:        %s" % name)
            self.out.output("VALUE:       %s" % param.value)
            self.out.output(
                wrapper.fill("DESCRIPTION: %s" % param.description))
            self.out.output()

    def do_sysprep(self):
        """Prepare system for image creation."""

        self.out.output('Preparing system for image creation:')

        if self.image.is_unsupported():
            self.out.warn(
                "System preparation is disabled for unsupported media")
            return

        enabled = [s for s in self.list_syspreps() if self.sysprep_enabled(s)]
        size = len(enabled)
        cnt = 0

        def exec_sysprep(cnt, size, task):
            self.out.output(('(%d/%d)' % (cnt, size)).ljust(7), False)
            task()
            del self._sysprep_tasks[task.__name__]

        with self.mount():
            for task in [t for t in enabled if t._sysprep_nomount is False]:
                cnt += 1
                exec_sysprep(cnt, size, task)

        for task in [t for t in enabled if t._sysprep_nomount]:
            cnt += 1
            exec_sysprep(cnt, size, task)

        self.out.output()

    @sysprep('Shrinking image (may take a while)', nomount=True)
    def _shrink(self):
        """Shrink the last file system and update the partition table"""
        self.image.shrink()
        self.shrinked = True

    @property
    def ismounted(self):
        return self._mounted

    def mount(self, readonly=False, silent=False, fatal=True):
        """Returns a context manager for mounting an image"""

        parent = self
        output = lambda msg='', nl=True: None if silent else self.out.output
        success = lambda msg='', nl=True: None if silent else self.out.success
        warn = lambda msg='', nl=True: None if silent else self.out.warn

        class Mount:
            """The Mount context manager"""
            def __enter__(self):
                mount_type = 'read-only' if readonly else 'read-write'
                output("Mounting the media %s ..." % mount_type, False)

                parent._mount_error = ""
                del parent._mount_warnings[:]

                try:
                    parent._mounted = parent._do_mount(readonly)
                except:
                    parent.image.g.umount_all()
                    raise

                if not parent.ismounted:
                    msg = "Unable to mount the media %s. Reason: %s" % \
                        (mount_type, parent._mount_error)
                    if fatal:
                        raise FatalError(msg)
                    else:
                        warn(msg)

                for warning in parent._mount_warnings:
                    warn(warning)

                if parent.ismounted:
                    success('done')

            def __exit__(self, exc_type, exc_value, traceback):
                output("Umounting the media ...", False)
                parent.image.g.umount_all()
                parent._mounted = False
                success('done')

        return Mount()

    def check_version(self, major, minor):
        """Checks the OS version against the one specified by the major, minor
        tuple.

        Returns:
            < 0 if the OS version is smaller than the specified one
            = 0 if they are equal
            > 0 if it is greater
        """
        guestfs = self.image.g
        for a, b in ((guestfs.inspect_get_major_version(self.root), major),
                     (guestfs.inspect_get_minor_version(self.root), minor)):
            if a != b:
                return a - b

        return 0

    @add_prefix
    def _ls(self, directory):
        """List the name of all files under a directory"""
        return self.image.g.ls(directory)

    @add_prefix
    def _find(self, directory):
        """List the name of all files recursively under a directory"""
        return self.image.g.find(directory)

    def _foreach_file(self, directory, action, **kargs):
        """Perform an action recursively on all files under a directory.

        The following options are allowed:

        * maxdepth: If defined, the action will not be performed on files that
          are below this level of directories under the directory parameter.

        * ftype: The action will only be performed on files of this type. For a
          list of all allowed file types, see here:
          http://libguestfs.org/guestfs.3.html#guestfs_readdir

        * exclude: Exclude all files that follow this pattern.
        """
        if not self.image.g.is_dir(directory):
            self.out.warn("Directory: `%s' does not exist!" % directory)
            return

        maxdepth = None if 'maxdepth' not in kargs else kargs['maxdepth']
        if maxdepth == 0:
            return

        # maxdepth -= 1
        maxdepth = None if maxdepth is None else maxdepth - 1
        kargs['maxdepth'] = maxdepth

        exclude = None if 'exclude' not in kargs else kargs['exclude']
        ftype = None if 'ftype' not in kargs else kargs['ftype']
        has_ftype = lambda x, y: y is None and True or x['ftyp'] == y

        for f in self.image.g.readdir(directory):
            if f['name'] in ('.', '..'):
                continue

            full_path = "%s/%s" % (directory, f['name'])

            if exclude and re.match(exclude, full_path):
                continue

            if has_ftype(f, 'd'):
                self._foreach_file(full_path, action, **kargs)

            if has_ftype(f, ftype):
                action(full_path)

    def _do_inspect(self):
        """helper method for inspect"""
        self.out.warn("No inspection method available")
        pass

    def _do_collect_metadata(self):
        """helper method for collect_metadata"""

        try:
            self.meta['ROOT_PARTITION'] = \
                "%d" % self.image.g.part_to_partnum(self.root)
        except RuntimeError:
            self.out.warn("Unable to identify the partition number from root "
                          "partition: %s" % self.root)

        self.meta['OSFAMILY'] = self.image.g.inspect_get_type(self.root)
        self.meta['OS'] = self.image.g.inspect_get_distro(self.root)
        if self.meta['OS'] == "unknown":
            self.meta['OS'] = self.meta['OSFAMILY']
        self.meta['DESCRIPTION'] = \
            self.image.g.inspect_get_product_name(self.root)

    def _do_mount(self, readonly):
        """helper method for mount"""
        try:
            self.image.g.mount_options(
                'ro' if readonly else 'rw', self.root, '/')
        except RuntimeError as msg:
            self._mount_error = str(msg)
            return False

        return True

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Normal Command-line interface output"""

from image_creator.output import Output

import sys
from colors import red, green, yellow
from progress.bar import Bar


def output(msg, new_line, decorate, stream):
    nl = "\n" if new_line else ' '
    stream.write(decorate(msg) + nl)


def error(msg, new_line, colored, stream):
    color = red if colored else lambda x: x
    output("Error: %s" % msg, new_line, color, stream)


def warn(msg, new_line, colored, stream):
    color = yellow if colored else lambda x: x
    output("Warning: %s" % msg, new_line, color, stream)


def success(msg, new_line, colored, stream):
    color = green if colored else lambda x: x
    output(msg, new_line, color, stream)


def clear(stream):
    """Clears the terminal screen."""
    if stream.isatty():
        stream.write('\033[H\033[2J')


class SilentOutput(Output):
    """Silent Output class. Only Errors are printed"""
    pass


class SimpleOutput(Output):
    """Print messages but not progress bars. Progress bars are treated as
    output messages. The user gets informed when the action begins and when it
    ends, but no progress is shown in between."""
    def __init__(self, colored=True, stream=None):
        self.colored = colored
        self.stream = sys.stderr if stream is None else stream

    def error(self, msg, new_line=True):
        """Print an error"""
        error(msg, new_line, self.colored, self.stream)

    def warn(self, msg, new_line=True):
        """Print a warning"""
        warn(msg, new_line, self.colored, self.stream)

    def success(self, msg, new_line=True):
        """Print msg after an action is completed"""
        success(msg, new_line, self.colored, self.stream)

    def output(self, msg='', new_line=True):
        """Print msg as normal program output"""
        output(msg, new_line, lambda x: x, self.stream)

    def clear(self):
        """Clear the screen"""
        clear(self.stream)


class OutputWthProgress(SimpleOutput):
    """Output class with progress."""
    class _Progress(Bar):
        MESSAGE_LENGTH = 30

        template = {
            'default': '%(index)d/%(max)d',
            'percent': '%(percent)d%%',
            'b': '%(index)d/%(max)d B',
            'kb': '%(index)d/%(max)d KB',
            'mb': '%(index)d/%(max)d MB'
        }

        def __init__(self, size, title, bar_type='default'):
            """Create a Progress bar"""
            self.hide_cursor = False
            super(OutputWthProgress._Progress, self).__init__()
            self.title = title
            self.fill = '#'
            self.bar_prefix = ' ['
            self.bar_suffix = '] '
            self.message = ("%s:" % self.title).ljust(self.MESSAGE_LENGTH)
            self.suffix = self.template[bar_type]
            self.max = size

            # print empty progress bar
            self.start()

        def success(self, result):
            """Print result after progress has finished"""
            self.output.output("\r%s ...\033[K" % self.title, False)
            self.output.success(result)

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module is the entrance point for the dialog-based version of the
snf-image-creator program. The main function will create a dialog where the
user is asked if he wants to use the program in expert or wizard mode.
"""

import dialog
import sys
import os
import textwrap
import signal
import optparse
import types
import termios
import traceback

from image_creator import __version__ as version
from image_creator.util import FatalError
from image_creator.output import Output
from image_creator.output.cli import SimpleOutput
from image_creator.output.dialog import GaugeOutput
from image_creator.output.composite import CompositeOutput
from image_creator.disk import Disk
from image_creator.dialog_wizard import start_wizard
from image_creator.dialog_menu import main_menu
from image_creator.dialog_util import WIDTH, confirm_exit, Reset, \
    update_background_title, select_file

PROGNAME = os.path.basename(sys.argv[0])


def create_image(d, media, out, tmp, snapshot):
    """Create an image out of `media'"""
    d.setBackgroundTitle('snf-image-creator')

    gauge = GaugeOutput(d, "Initialization", "Initializing...")
    out.add(gauge)
    disk = Disk(media, out, tmp)

    def signal_handler(signum, frame):
        gauge.cleanup()
        disk.cleanup()

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    try:

        device = disk.file if not snapshot else disk.snapshot()

        image = disk.get_image(device)

        out.output("Collecting image metadata ...")
        metadata = {}
        for (key, value) in image.meta.items():
            metadata[str(key)] = str(value)

        for (key, value) in image.os.meta.items():
            metadata[str(key)] = str(value)

        out.success("done")
        gauge.cleanup()
        out.remove(gauge)

        # Make sure the signal handler does not call gauge.cleanup again
        def dummy(self):
            pass
        gauge.cleanup = type(GaugeOutput.cleanup)(dummy, gauge, GaugeOutput)

        session = {"dialog": d,
                   "disk": disk,
                   "image": image,
                   "metadata": metadata}

        if image.is_unsupported():

            session['excluded_tasks'] = [-1]
            session['task_metadata'] = ["EXCLUDE_ALL_TASKS"]

            msg = "The system on the input media is not supported." \
                "\n\nReason: %s\n\n" \
                "We highly recommend not to create an image out of this, " \
                "since the image won't be cleaned up and you will not be " \
                "able to configure it during the deployment. Press <YES> if " \
                "you still want to continue with the image creation process." \
                % image._unsupported

            if not d.yesno(msg, width=WIDTH, defaultno=1, height=12):
                main_menu(session)

            d.infobox("Thank you for using snf-image-creator. Bye", width=53)
            return 0

        msg = "snf-image-creator detected a %s system on the input media. " \
              "Would you like to run a wizard to assist you through the " \
              "image creation process?\n\nChoose <Wizard> to run the wizard," \
              " <Expert> to run snf-image-creator in expert mode or press " \
              "ESC to quit the program." \
              % (image.ostype.capitalize() if image.ostype == image.distro or
                 image.distro == "unknown" else "%s (%s)" %
                 (image.ostype.capitalize(), image.distro.capitalize()))

        update_background_title(session)

        while True:
            code = d.yesno(msg, width=WIDTH, height=12, yes_label="Wizard",
                           no_label="Expert")
            if code == d.DIALOG_OK:
                if start_wizard(session):
                    break
            elif code == d.DIALOG_CANCEL:
                main_menu(session)
                break

            if confirm_exit(d):
                break

        d.infobox("Thank you for using snf-image-creator. Bye", width=53)
    finally:
        disk.cleanup()

    return 0


def _dialog_form(self, text, height=20, width=60, form_height=15, fields=[],
                 **kwargs):
    """Display a form box.

    fields is in the form: [(label1, item1, item_length1), ...]
    """

    cmd = ["--form", text, str(height), str(width), str(form_height)]

    label_len = 0
    for field in fields:
        if len(field[0]) > label_len:
            label_len = len(field[0])

    input_len = width - label_len - 1

    line = 1
    for field in fields:
        label = field[0]
        item = field[1]
        item_len = field[2]
        cmd.extend((label, str(line), str(1), item, str(line),
                    str(label_len + 1), str(input_len), str(item_len)))
        line += 1

    code, output = self._perform(*(cmd,), **kwargs)

    if not output:
        return (code, [])

    return (code, output.splitlines())


def dialog_main(media, logfile, tmpdir, snapshot):

    # In openSUSE dialog is buggy under xterm
    if os.environ['TERM'] == 'xterm':
        os.environ['TERM'] = 'linux'

    d = dialog.Dialog(dialog="dialog")

    # Add extra button in dialog library
    dialog._common_args_syntax["extra_button"] = \
        lambda enable: dialog._simple_option("--extra-button", enable)
    dialog._common_args_syntax["extra_label"] = \
        lambda string: ("--extra-label", string)

    # Allow yes-no label overwriting
    dialog._common_args_syntax["yes_label"] = \
        lambda string: ("--yes-label", string)
    dialog._common_args_syntax["no_label"] = \
        lambda string: ("--no-label", string)

    # Monkey-patch pythondialog to include support for form dialog boxes
    if not hasattr(dialog, 'form'):
        d.form = types.MethodType(_dialog_form, d)

    d.setBackgroundTitle('snf-image-creator')

    try:
        while True:
            media = select_file(d, init=media, ftype="br", bundle_host=True,
                                title="Please select an input media.")
            if media is None:
                if confirm_exit(
                        d, "You canceled the media selection dialog box."):
                    return 0
                continue
            break

        log = SimpleOutput(False, logfile) if logfile is not None else Output()
        while 1:
            try:
                out = CompositeOutput([log])
                out.output("Starting %s v%s ..." % (PROGNAME, version))
                return create_image(d, media, out, tmpdir, snapshot)
            except Reset:
                log.output("Resetting everything ...")
                continue
    except FatalError as error:
        msg = textwrap.fill(str(error), width=WIDTH-4)
        d.infobox(msg, width=WIDTH, title="Fatal Error")
        return 1


def main():
    """Entrance Point"""
    if os.geteuid() != 0:
        sys.stderr.write("Error: You must run %s as root\n" % PROGNAME)
        sys.exit(2)

    usage = "Usage: %prog [options] [<input_media>]"
    parser = optparse.OptionParser(version=version, usage=usage)
    parser.add_option("-l", "--logfile", type="string", dest="logfile",
                      default=None, help="log all messages to FILE",
                      metavar="FILE")
    parser.add_option("--no-snapshot", dest="snapshot", default=True,
                      help="don't snapshot the input media. (THIS IS "
                      "DANGEROUS AS IT WILL ALTER THE ORIGINAL MEDIA!!!)",
                      action="store_false")
    parser.add_option("--tmpdir", type="string", dest="tmp", default=None,
                      help="create large temporary image files under DIR",
                      metavar="DIR")

    opts, args = parser.parse_args(sys.argv[1:])

    if len(args) > 1:
        parser.error("Wrong number of arguments")

    media = args[0] if len(args) == 1 else None

    if opts.tmp is not None and not os.path.isdir(opts.tmp):
        parser.error("Directory: `%s' specified with --tmpdir is not valid"
                     % opts.tmp)

    try:
        logfile = open(opts.logfile, 'w') if opts.logfile is not None else None
    except IOError as error:
        parser.error("Unable to open logfile `%s' for writing. Reason: %s" %
                     (opts.logfile, error.strerror))

    try:
        # Save the terminal attributes
        attr = termios.tcgetattr(sys.stdin.fileno())
        try:
            ret = dialog_main(media, logfile, opts.tmp, opts.snapshot)
        finally:
            # Restore the terminal attributes. If an error occurs make sure
            # that the terminal turns back to normal.
            termios.tcsetattr(sys.stdin.fileno(), termios.TCSADRAIN, attr)
    except:
        # Clear the screen
        sys.stdout.write('\033[2J')  # Erase Screen
        sys.stdout.write('\033[H')  # Cursor Home
        sys.stdout.flush()

        exception = traceback.format_exc()
        sys.stderr.write(exception)
        if logfile is not None:
            logfile.write(exception)

        sys.exit(3)
    finally:
        if logfile is not None:
            logfile.close()

    sys.exit(ret)

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module hosts OS-specific code for FreeBSD."""

from image_creator.os_type.unix import Unix, sysprep

import re


class Freebsd(Unix):
    """OS class for FreeBSD Unix-like operating system"""

    @sysprep("Cleaning up passwords & locking all user accounts")
    def _cleanup_password(self):
        """Remove all passwords and lock all user accounts"""

        master_passwd = []

        for line in self.image.g.cat('/etc/master.passwd').splitlines():

            # Check for empty or comment lines
            if len(line.split('#')[0]) == 0:
                master_passwd.append(line)
                continue

            fields = line.split(':')
            if fields[1] not in ('*', '!'):
                fields[1] = '!'

            master_passwd.append(":".join(fields))

        self.image.g.write(
            '/etc/master.passwd', "\n".join(master_passwd) + '\n')

        # Make sure no one can login on the system
        self.image.g.rm_rf('/etc/spwd.db')

    def _do_collect_metadata(self):
        """Collect metadata about the OS"""
        super(Freebsd, self)._do_collect_metadata()
        self.meta["USERS"] = " ".join(self._get_passworded_users())

        # The original product name key is long and ugly
        self.meta['DESCRIPTION'] = \
            self.meta['DESCRIPTION'].split('#')[0].strip()

        # Delete the USERS metadata if empty
        if not len(self.meta['USERS']):
            self.out.warn("No passworded users found!")
            del self.meta['USERS']

    def _do_inspect(self):
        """Run various diagnostics to check if media is supported"""

        self.out.output('Checking partition table type...', False)
        ptype = self.image.g.part_get_parttype(self.image.guestfs_device)
        if ptype != 'gpt':
            self.out.warn("partition table type is: `%s'" % ptype)
            self.image.set_unsupported(
                'On FreeBSD only GUID partition tables are supported')
        else:
            self.out.success(ptype)

    def _get_passworded_users(self):
        """Returns a list of non-locked user accounts"""
        users = []
        regexp = re.compile(
            '^([^:]+):((?:![^:]+)|(?:[^!*][^:]+)|):(?:[^:]*:){7}(?:[^:]*)'
        )

        for line in self.image.g.cat('/etc/master.passwd').splitlines():
            line = line.split('#')[0]
            match = regexp.match(line)
            if not match:
                continue

            user, passwd = match.groups()
            if len(passwd) > 0 and passwd[0] == '!':
                self.out.warn("Ignoring locked %s account." % user)
            else:
                users.append(user)

        return users

    def _do_mount(self, readonly):
        """Mount partitions in the correct order"""

        critical_mpoints = ('/', '/etc', '/root', '/home', '/var')

        # Older libguestfs versions can't handle correct FreeBSD partitions on
        # a GUID Partition Table. We have to do the translation to Linux device
        # names ourselves
        guid_device = re.compile(r'^/dev/((?:ada)|(?:vtbd))(\d+)p(\d+)$')

        mopts = "ufstype=ufs2,%s" % ('ro' if readonly else 'rw')
        for mp, dev in self._mountpoints():
            match = guid_device.match(dev)
            if match:
                group2 = int(match.group(2))
                group3 = int(match.group(3))
                dev = '/dev/sd%c%d' % (chr(ord('a') + group2), group3)
            try:
                self.image.g.mount_vfs(mopts, 'ufs', dev, mp)
            except RuntimeError as msg:
                if mp in critical_mpoints:
                    self._mount_error = str(msg)
                    return False
                else:
                    self._mount_warnings.append('%s (ignored)' % msg)

        return True

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module implements the "wizard" mode of the dialog-based version of
snf-image-creator.
"""

import time
import StringIO
import json
import re

from image_creator.kamaki_wrapper import Kamaki, ClientError
from image_creator.util import FatalError, virtio_versions
from image_creator.output.cli import OutputWthProgress
from image_creator.dialog_util import extract_image, update_background_title, \
    add_cloud, edit_cloud, update_sysprep_param

PAGE_WIDTH = 70
PAGE_HEIGHT = 12


class WizardExit(Exception):
    """Exception used to exit the wizard"""
    pass


class WizardReloadPage(Exception):
    """Exception that reloads the last WizardPage"""
    pass


class Wizard(object):
    """Represents a dialog-based wizard

    The wizard is a collection of pages that have a "Next" and a "Back" button
    on them. The pages are used to collect user data.
    """

    def __init__(self, dialog):
        """Initialize the Wizard"""
        self._pages = []
        self.dialog = dialog

    def add_page(self, page):
        """Add a new page to the wizard"""
        self._pages.append(page)

    def run(self):
        """Run the wizard"""
        idx = 0
        while True:
            try:
                total = len(self._pages)
                title = "(%d/%d) %s" % (idx + 1, total, self._pages[idx].title)
                idx += self._pages[idx].show(self.dialog, title)
            except WizardExit:
                return False
            except WizardReloadPage:
                continue

            if idx >= len(self._pages):
                text = "All necessary information has been gathered:\n\n"
                for page in self._pages:
                    text += " * %s\n" % page
                text += "\nContinue with the image creation process?"

                ret = self.dialog.yesno(
                    text, width=PAGE_WIDTH, height=8 + len(self._pages),
                    ok_label="Yes", cancel="Back", extra_button=1,
                    extra_label="Quit", title="Confirmation")

                if ret == self.dialog.DIALOG_CANCEL:
                    idx -= 1
                elif ret == self.dialog.DIALOG_EXTRA:
                    return False
                elif ret == self.dialog.DIALOG_OK:
                    return True

            if idx < 0:
                return False

    @property
    def answers(self):
        """Returns the answers the user provided"""
        return dict((page.name, page.answer) for page in self._pages)


class WizardPage(object):
    """Represents a page in a wizard"""
    NEXT = 1
    PREV = -1

    def __init__(self, name, text, **kargs):
        self.name = name
        self.answer = None
        self.text = text
        self.print_name = kargs['print_name'] if 'print_name' in kargs \
            else " ".join(re.findall('[A-Z][^A-Z]*', name))
        self.title = kargs['title'] if 'title' in kargs else self.print_name
        self.default = kargs['default'] if 'default' in kargs else ""
        self.extra = kargs['extra'] if 'extra' in kargs else None
        self.validate = \
            kargs['validate'] if 'validate' in kargs else lambda x: x
        self.display = kargs['display'] if 'display' in kargs else lambda x: x

        self.dargs = {}
        self.dargs['ok_label'] = 'Next'
        self.dargs['cancel'] = 'Back'
        self.dargs['width'] = PAGE_WIDTH
        self.dargs['height'] = PAGE_HEIGHT

        if 'extra' in kargs:
            self.dargs['extra_button'] = 1

        self.extra_label = kargs['extra_label'] if 'extra_label' in kargs \
            else lambda: "extra"

    def __str__(self):
        """Prints the answer"""
        return "%s: %s" % (self.print_name, self.display(self.answer))

    def show(self, dialog, title):
        """Display this wizard page

        This function is used by the wizard program when accessing a page.
        """
        raise NotImplementedError


class WizardInputPage(WizardPage):
    """Represents an input field in a wizard"""

    def show(self, dialog, title):
        """Display this wizard page"""
        (code, answer) = dialog.inputbox(self.text(), init=self.default,
                                         title=title,
                                         extra_label=self.extra_label(),
                                         **self.dargs)

        if code in (dialog.DIALOG_CANCEL, dialog.DIALOG_ESC):
            return self.PREV

        self.answer = self.validate(answer.strip())
        self.default = self.answer

        return self.NEXT


class WizardInfoPage(WizardPage):
    """Represents a Wizard Page that just displays some user-defined
    information.

    The user-defined information is created by the info function.
    """
    def __init__(self, name, text, info, **kargs):
        """Initialize the WizardInfoPage instance"""
        super(WizardInfoPage, self).__init__(name, text, **kargs)
        self.info = info

    def show(self, dialog, title):
        """Display this wizard page"""

        text = "%s\n\n%s" % (self.text(), self.info())

        ret = dialog.yesno(text, title=title, extra_label=self.extra_label(),
                           **self.dargs)

        if ret in (dialog.DIALOG_CANCEL, dialog.DIALOG_ESC):
            return self.PREV
        elif ret == dialog.DIALOG_EXTRA:
            self.extra()
            raise WizardReloadPage

        # DIALOG_OK
        self.answer = self.validate(None)
        return self.NEXT


class WizardFormPage(WizardPage):
    """Represents a Form in a wizard"""

    def __init__(self, name, text, fields, **kargs):
        """Initialize the WizardFormPage instance"""
        super(WizardFormPage, self).__init__(name, text, **kargs)
        self.fields = fields

    def show(self, dialog, title):
        """Display this wizard page"""
        field_lenght = len(self.fields())
        form_height = field_lenght if field_lenght < PAGE_HEIGHT - 4 \
            else PAGE_HEIGHT - 4

        (code, output) = dialog.form(self.text(), form_height=form_height,
                                     fields=self.fields(), title=title,
                                     extra_label=self.extra_label(),
                                     default_item=self.default, **self.dargs)

        if code in (dialog.DIALOG_CANCEL, dialog.DIALOG_ESC):
            return self.PREV

        self.answer = self.validate(output)
        self.default = output

        return self.NEXT


class WizardPageWthChoices(WizardPage):
    """Represents a Wizard Page that allows the user to select something from
    a list of choices.

    The available choices are created by a function passed to the class through
    the choices variable. If the choices function returns an empty list, a
    fallback function is executed if available.
    """
    def __init__(self, name, text, choices, **kargs):
        """Initialize the WizardPageWthChoices instance"""
        super(WizardPageWthChoices, self).__init__(name, text, **kargs)
        self.choices = choices
        self.fallback = kargs['fallback'] if 'fallback' in kargs else None


class WizardRadioListPage(WizardPageWthChoices):
    """Represent a Radio List in a wizard"""

    def show(self, dialog, title):
        """Display this wizard page"""
        choices = []
        for choice in self.choices():
            default = 1 if choice[0] == self.default else 0
            choices.append((choice[0], choice[1], default))

        (code, answer) = dialog.radiolist(self.text(), choices=choices,
                                          extra_label=self.extra_label(),
                                          title=title, **self.dargs)

        if code in (dialog.DIALOG_CANCEL, dialog.DIALOG_ESC):
            return self.PREV

        self.answer = self.validate(answer)
        self.default = answer

        return self.NEXT


class WizardMenuPage(WizardPageWthChoices):
    """Represents a menu dialog with available choices in a wizard"""

    def show(self, dialog, title):
        """Display this wizard page"""

        choices = self.choices()

        if len(choices) == 0:
            assert self.fallback, "Zero choices and no fallback"
            if self.fallback():
                raise WizardReloadPage
            else:
                return self.PREV

        default_item = self.default if self.default else choices[0][0]

        (code, choice) = dialog.menu(self.text(), title=title, choices=choices,
                                     extra_label=self.extra_label(),
                                     default_item=default_item, **self.dargs)

        if code in (dialog.DIALOG_CANCEL, dialog.DIALOG_ESC):
            return self.PREV
        elif code == dialog.DIALOG_EXTRA:
            self.extra()
            raise WizardReloadPage

        self.answer = self.validate(choice)
        self.default = choice

        return self.NEXT


def start_wizard(session):
    """Run the image creation wizard"""

    metadata = session['metadata']
    distro = session['image'].distro
    ostype = session['image'].ostype

    # Create Cloud Wizard Page
    def cloud_choices():
        """Returns the available clouds"""
        choices = []
        for (name, cloud) in Kamaki.get_clouds().items():
            descr = cloud['description'] if 'description' in cloud else ''
            choices.append((name, descr))

        return choices

    def no_clouds():
        """Fallback function when no cloud account exists"""
        if not session['dialog'].yesno(
                "No available clouds found. Would you like to add one now?",
                width=PAGE_WIDTH, defaultno=0):
            return add_cloud(session)
        return False

    def cloud_validate(cloud):
        """Checks if a cloud is valid"""
        if not Kamaki.get_account(cloud):
            if not session['dialog'].yesno(
                    "The cloud you have selected is not valid! Would you "
                    "like to edit it now?", width=PAGE_WIDTH, defaultno=0):
                if edit_cloud(session, cloud):
                    return cloud
            raise WizardReloadPage
        return cloud

    cloud = WizardMenuPage(
        "Cloud", lambda:
        "Please select a cloud account or press <Add> to add a new one:",
        cloud_choices, extra_label=lambda: "Add",
        extra=lambda: add_cloud(session), title="Clouds",
        validate=cloud_validate, fallback=no_clouds)

    # Create Image Name Wizard Page
    name = WizardInputPage("ImageName", lambda:
                           "Please provide a name for the image:",
                           default=ostype if distro == "unknown" else distro)

    # Create Image Description Wizard Page
    descr = WizardInputPage(
        "ImageDescription", lambda:
        "Please provide a description for the image:",
        default=metadata['DESCRIPTION'] if 'DESCRIPTION' in metadata else '')

    # Create VirtIO Installation Page
    def display_installed_drivers():
        """Returns the installed VirtIO drivers"""
        image = session['image']
        versions = virtio_versions(image.os.virtio_state)

        ret = "Installed Block Device Driver:  %(netkvm)s\n" \
              "Installed Network Device Driver: %(viostor)s\n" % versions

        virtio = image.os.sysprep_params['virtio'].value
        if virtio:
            ret += "\nBlock Device Driver to be installed:   %(netkvm)s\n" \
                   "Network Device Driver to be installed: %(viostor)s\n" % \
                   virtio_versions(image.os.compute_virtio_state(virtio))
        return ret

    def validate_virtio(_):
        """Checks the state of the VirtIO drivers"""
        image = session['image']
        netkvm = len(image.os.virtio_state['netkvm']) != 0
        viostor = len(image.os.virtio_state['viostor']) != 0
        drv_dir = image.os.sysprep_params['virtio'].value

        if netkvm is False or viostor is False:
            new = image.os.compute_virtio_state(drv_dir) if drv_dir else None
            new_viostor = len(new['viostor']) != 0 if new else False
            new_netkvm = len(new['netkvm']) != 0 if new else False

            dialog = session['dialog']
            title = "VirtIO driver missing"
            msg = "Image creation cannot proceed unless a VirtIO %s driver " \
                  "is installed on the media!"
            if not (viostor or new_viostor):
                dialog.msgbox(msg % "Block Device", width=PAGE_WIDTH,
                              height=PAGE_HEIGHT, title=title)
                raise WizardReloadPage
            if not(netkvm or new_netkvm):
                dialog.msgbox(msg % "Network Device", width=PAGE_WIDTH,
                              height=PAGE_HEIGHT, title=title)
                raise WizardReloadPage

        return drv_dir

    def virtio_text():
        if not session['image'].os.sysprep_params['virtio'].value:
            return "Press <New> to update the image's VirtIO drivers."
        else:
            return "Press <Revert> to revert to the old state."

    def virtio_extra():
        if not session['image'].os.sysprep_params['virtio'].value:
            title = "Please select a directory that hosts VirtIO drivers."
            update_sysprep_param(session, 'virtio', title=title)
        else:
            session['image'].os.sysprep_params['virtio'].value = ""

    def virtio_extra_label():
        if not session['image'].os.sysprep_params['virtio'].value:
            return "New"
        else:
            return "Revert"

    virtio = WizardInfoPage(
        "virtio", virtio_text, display_installed_drivers,
        title="VirtIO Drivers", extra_label=virtio_extra_label,
        extra=virtio_extra, validate=validate_virtio,
        print_name="VirtIO Drivers Path")

    # Create Image Registration Wizard Page
    def registration_choices():
        """Choices for the registration wizard page"""
        return [("Private", "Image is accessible only by this user"),
                ("Public", "Everyone can create VMs from this image")]

    registration = WizardRadioListPage("RegistrationType", lambda:
                                       "Please provide a registration type:",
                                       registration_choices, default="Private")

    wizard = Wizard(session['dialog'])

    wizard.add_page(cloud)
    wizard.add_page(name)
    wizard.add_page(descr)
    if hasattr(session['image'].os, 'install_virtio_drivers'):
        wizard.add_page(virtio)
    wizard.add_page(registration)

    if wizard.run():
        create_image(session, wizard.answers)
    else:
        return False

    return True


def create_image(session, answers):
    """Create an image using the information collected by the wizard"""
    image = session['image']

    with_progress = OutputWthProgress(True)
    image.out.add(with_progress)
    try:
        image.out.clear()

        if 'virtio' in answers and image.os.sysprep_params['virtio'].value:
            image.os.install_virtio_drivers()

        # Sysprep
        image.os.do_sysprep()
        metadata = image.os.meta

        update_background_title(session)

        metadata.update(image.meta)
        metadata['DESCRIPTION'] = answers['ImageDescription']

        # MD5
        session['checksum'] = image.md5()

        image.out.output()
        try:
            image.out.output("Uploading image to the cloud:")
            account = Kamaki.get_account(answers['Cloud'])
            assert account, "Cloud: %s is not valid" % answers['Cloud']
            kamaki = Kamaki(account, image.out)

            name = "%s-%s.diskdump" % (answers['ImageName'],
                                       time.strftime("%Y%m%d%H%M"))
            with image.raw_device() as raw:
                with open(raw, 'rb') as device:
                    remote = kamaki.upload(device, image.size, name,
                                           "(1/3)  Calculating block hashes",
                                           "(2/3)  Uploading image blocks")

            image.out.output("(3/3)  Uploading md5sum file ...", False)
            md5sumstr = '%s %s\n' % (session['checksum'], name)
            kamaki.upload(StringIO.StringIO(md5sumstr), size=len(md5sumstr),
                          remote_path="%s.%s" % (name, 'md5sum'))
            image.out.success('done')
            image.out.output()

            image.out.output('Registering %s image with the cloud ...' %
                             answers['RegistrationType'].lower(), False)
            result = kamaki.register(answers['ImageName'], remote, metadata,
                                     answers['RegistrationType'] == "Public")
            image.out.success('done')
            image.out.output("Uploading metadata file ...", False)
            metastring = unicode(json.dumps(result, ensure_ascii=False))
            kamaki.upload(StringIO.StringIO(metastring), size=len(metastring),
                          remote_path="%s.%s" % (name, 'meta'))
            image.out.success('done')

            if answers['RegistrationType'] == "Public":
                image.out.output("Sharing md5sum file ...", False)
                kamaki.share("%s.md5sum" % name)
                image.out.success('done')
                image.out.output("Sharing metadata file ...", False)
                kamaki.share("%s.meta" % name)
                image.out.success('done')

            image.out.output()

        except ClientError as error:
            raise FatalError("Storage service client: %d %s" %
                             (error.status, error.message))
    finally:
        image.out.remove(with_progress)

    text = "The %s image was successfully uploaded to the storage service " \
           "and registered with the compute service of %s. Would you like " \
           "to keep a local copy?" % \
           (answers['RegistrationType'].lower(), answers['Cloud'])

    if not session['dialog'].yesno(text, width=PAGE_WIDTH):
        extract_image(session)

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module hosts Windows PowerShell scripts that need to be injected into
the windows image"""

# Just a random 16 character long token
from image_creator.os_type.windows.vm import RANDOM_TOKEN

COM1_WRITE = r"""
$port=new-Object System.IO.Ports.SerialPort COM1,9600,None,8,one
$port.open()
$port.WriteLine('""" + RANDOM_TOKEN + """')
$port.Close()

"""

# Installs drivers found in a directory
DRVINST_HEAD = r"""
#requires -version 2

Param([string]$dirName=$(throw "You need to provide a directory name"))

if (!(Test-Path -PathType Container "$dirName")) {
    Write-Error -Category InvalidArgument "Invalid Directory: $dirName"
    Exit
}

""" + COM1_WRITE + r"""

function Cat2Cert
{
    $catFile = Get-Item $args[0]
    $cert = (Get-AuthenticodeSignature $catFile.FullName).SignerCertificate
    $certFile = "$dirName" + "\" + $catFile.BaseName + ".cer"
    [System.IO.File]::WriteAllBytes($certFile, $cert.Export("Cert"))
    return $certFile
}

# This is not needed on VMs
powercfg.exe /hibernate off
"""

ADD_CERTIFICATE = r"""
CertUtil -addstore TrustedPublisher "$(Cat2Cert "$dirName\%s")"
"""

ADD_DRIVER = r"""
PnPUtil -a "$dirName\%s"
"""

INSTALL_DRIVER = r"""
PnPUtil -i -a "$dirName\%s"
"""

REMOVE_DRIVER = r"""
PnPUtil -f -d %s
"""

DRVINST_TAIL = COM1_WRITE + """
shutdown /s /t 0
"""

DISABLE_AUTOLOGON = r"""
Remove-ItemProperty -Path `
    'HKLM:\Software\Microsoft\Windows NT\CurrentVersion\Winlogon\' `
    DefaultUserName
Remove-ItemProperty -Path `
    'HKLM:\Software\Microsoft\Windows NT\CurrentVersion\Winlogon\' `
    DefaultPassword
Remove-ItemProperty -Path `
    'HKLM:\Software\Microsoft\Windows NT\CurrentVersion\Winlogon\' `
    AutoAdminLogon
"""

# Reboots system in safe mode
SAFEBOOT = r"""
bcdedit /set safeboot minimal

$winlogon = 'HKLM\SOFTWARE\Microsoft\Windows NT\CurrentVersion\Winlogon'
New-ItemProperty `
    -Path HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion\RunOnce `
    -Name *1snf-image-creator-safeboot -PropertyType String `
    -Value 'bcdedit /deletevalue safeboot'

New-ItemProperty `
    -Path HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion\RunOnce `
    -Name *2snf-image-creator-safeboot -PropertyType String `
    -Value "reg delete `"$winlogon`" /v DefaultUserName /f"
New-ItemProperty `
    -Path HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion\RunOnce `
    -Name *3snf-image-creator-safeboot -PropertyType String `
    -Value "reg delete `"$winlogon`" /v DefaultPassword /f"
New-ItemProperty `
    -Path HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion\RunOnce `
    -Name *4snf-image-creator-safeboot -PropertyType String `
    -Value "reg delete `"$winlogon`" /v AutoAdminLogon /f"

New-ItemProperty `
    -Path HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion\RunOnce `
    -Name *5snf-image-creator-safeboot -PropertyType String `
    -Value 'shutdown /s /t 5'
"""

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This package hosts code for accessing the windows registry"""

from image_creator.util import FatalError

import hivex
import tempfile
import os
import struct

# The Administrators group RID
ADMINS = 0x00000220

# http://technet.microsoft.com/en-us/library/hh824815.aspx
WINDOWS_SETUP_STATES = (
    "IMAGE_STATE_COMPLETE",
    "IMAGE_STATE_UNDEPLOYABLE",
    "IMAGE_STATE_GENERALIZE_RESEAL_TO_OOBE",
    "IMAGE_STATE_GENERALIZE_RESEAL_TO_AUDIT",
    "IMAGE_STATE_SPECIALIZE_RESEAL_TO_OOBE",
    "IMAGE_STATE_SPECIALIZE_RESEAL_TO_AUDIT")

REG_SZ = lambda k, v: {'key': k, 't': 1L,
                       'value': (v + '\x00').encode('utf-16le')}
REG_EXPAND_SZ = lambda k, v: {'key': k, 't': 2L,
                              'value': (v + '\x00').encode('utf-16le')}
REG_BINARY = lambda k, v: {'key': k, 't': 3L, 'value': v}
REG_DWORD = lambda k, v: {'key': k, 't': 4L, 'value': struct.pack('<I', v)}


def safe_add_node(hive, parent, name):
    """Add a registry node only if it is not present"""

    node = hive.node_get_child(parent, name)
    return hive.node_add_child(parent, name) if node is None else node


def traverse(hive, path):
    """Traverse a hive following a path"""

    node = hive.root()

    for name in path.split('/'):

        if len(name) == 0:
            continue

        node = hive.node_get_child(node, name)
        if node is None:
            break

    return node


class Registry(object):
    """Windows Registry manipulation methods"""

    def __init__(self, image):
        # Do not copy the guestfs handler. It may be overwritten by the image
        # class in the future
        self.image = image
        self.root = image.root

    def open_hive(self, hive, write=False):
        """Returns a context manager for opening a hive file of the image for
        reading or writing.
        """
        systemroot = self.image.g.inspect_get_windows_systemroot(self.root)
        path = "%s/system32/config/%s" % (systemroot, hive)
        try:
            path = self.image.g.case_sensitive_path(path)
        except RuntimeError as err:
            raise FatalError("Unable to retrieve file: %s. Reason: %s" %
                             (hive, str(err)))

        # OpenHive class needs this since 'self' gets overwritten
        g = self.image.g

        class OpenHive:
            """The OpenHive context manager"""
            def __enter__(self):
                localfd, self.localpath = tempfile.mkstemp()
                try:
                    os.close(localfd)
                    g.download(path, self.localpath)

                    hive = hivex.Hivex(self.localpath, write=write)
                except:
                    os.unlink(self.localpath)
                    raise

                return hive

            def __exit__(self, exc_type, exc_value, traceback):
                try:
                    if write:
                        g.upload(self.localpath, path)
                finally:
                    os.unlink(self.localpath)

        return OpenHive()

    @property
    def current_control_set(self):
        """Returns the current control set of the registry"""

        if hasattr(self, '_current_control_set'):
            return self._current_control_set

        with self.open_hive('SYSTEM') as hive:
            select = hive.node_get_child(hive.root(), 'Select')
            current_value = hive.node_get_value(select, 'Current')

            # expecting a little endian dword
            assert hive.value_type(current_value)[1] == 4
            current = "%03d" % hive.value_dword(current_value)

            self._current_control_set = 'ControlSet%s' % current

        return self._current_control_set

    def get_setup_state(self):
        """Returns the stage of Windows Setup the image is in.
        The method will return an int with one of the following values:
            0 => IMAGE_STATE_COMPLETE
            1 => IMAGE_STATE_GENERALIZE_RESEAL_TO_OOBE
            2 => IMAGE_STATE_GENERALIZE_RESEAL_TO_AUDIT
            3 => IMAGE_STATE_SPECIALIZE_RESEAL_TO_OOBE
            4 => IMAGE_STATE_SPECIALIZE_RESEAL_TO_AUDIT
        For more information check here:
            http://technet.microsoft.com/en-us/library/hh824815.aspx
        """

        with self.open_hive('SOFTWARE') as hive:
            # Navigate to:
            #   SOFTWARE/Microsoft/Windows/CurrentVersion/Setup/State
            state = hive.root()
            for child in ('Microsoft', 'Windows', 'CurrentVersion',
                          'Setup', 'State'):
                state = hive.node_get_child(state, child)

            image_state = hive.node_get_value(state, 'ImageState')
            vtype, value = hive.value_value(image_state)
            assert vtype == 1L, \
                "ImageState field type (=%d) is not REG_SZ" % vtype

            value = value.decode('utf-16le')

            ret = 0
            for known_state in WINDOWS_SETUP_STATES:
                if value == known_state + '\x00':  # REG_SZ is null-terminated
                    return ret
                ret += 1

        raise FatalError("Unknown Windows Setup State: %s" % value)

    def runonce(self, commands):
        """Add commands to the RunOnce registry key"""

        with self.open_hive('SOFTWARE', write=True) as hive:

            key = hive.root()
            for child in ('Microsoft', 'Windows', 'CurrentVersion'):
                key = hive.node_get_child(key, child)

            runonce = hive.node_get_child(key, "RunOnce")
            if runonce is None:
                runonce = hive.node_add_child(key, "RunOnce")

            for desc, cmd in commands.items():
                hive.node_set_value(runonce, REG_SZ(desc, cmd))

            hive.commit(None)

    def enable_autologon(self, username, password=""):
        """Enable automatic logon for a specific user"""

        with self.open_hive('SOFTWARE', write=True) as hive:

            winlogon = hive.root()
            for child in ('Microsoft', 'Windows NT', 'CurrentVersion',
                          'Winlogon'):
                winlogon = hive.node_get_child(winlogon, child)

            hive.node_set_value(winlogon, REG_SZ('DefaultUserName', username))
            hive.node_set_value(winlogon, REG_SZ('DefaultPassword', password))
            hive.node_set_value(winlogon, REG_SZ('AutoAdminLogon', "1"))

            hive.commit(None)

    def update_firewalls(self, domain, public, standard):
        """Enables or disables the firewall for the Domain, the Public and the
        Standard profile. Returns a triple with the old values.

        1 will enable a firewall and 0 will disable it
        """

        if domain not in (0, 1):
            raise ValueError("Valid values for domain parameter are 0 and 1")

        if public not in (0, 1):
            raise ValueError("Valid values for public parameter are 0 and 1")

        if standard not in (0, 1):
            raise ValueError("Valid values for standard parameter are 0 and 1")

        with self.open_hive('SYSTEM', write=True) as hive:
            firewall_policy = hive.root()
            for child in (self.current_control_set, 'services', 'SharedAccess',
                          'Parameters', 'FirewallPolicy'):
                firewall_policy = hive.node_get_child(firewall_policy, child)

            old_values = []
            new_values = [domain, public, standard]
            for profile in ('Domain', 'Public', 'Standard'):
                node = hive.node_get_child(firewall_policy,
                                           '%sProfile' % profile)

                old_value = hive.node_get_value(node, 'EnableFirewall')

                # expecting a little endian dword
                assert hive.value_type(old_value)[1] == 4
                old_values.append(hive.value_dword(old_value))

                hive.node_set_value(node, REG_DWORD('EnableFirewall',
                                                    new_values.pop(0)))
            hive.commit(None)

        return old_values

    def update_uac_remote_setting(self, value):
        r"""Updates the registry key value:
        [HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Policies
        \System]"LocalAccountTokenFilterPolicy"

        value = 1 will disable the UAC remote restrictions
        value = 0 will enable the UAC remote restrictions

        For more info see here: http://support.microsoft.com/kb/951016

        Returns:
            True if the key is changed
            False if the key is unchanged
        """

        if value not in (0, 1):
            raise ValueError("Valid values for value parameter are 0 and 1")

        with self.open_hive('SOFTWARE', write=True) as hive:
            path = 'Microsoft/Windows/CurrentVersion/Policies/System'
            system = traverse(hive, path)

            policy = None
            for val in hive.node_values(system):
                if hive.value_key(val) == "LocalAccountTokenFilterPolicy":
                    policy = val

            if policy is not None:
                if value == hive.value_dword(policy):
                    return False
            elif value == 0:
                return False

            hive.node_set_value(
                system, REG_DWORD("LocalAccountTokenFilterPolicy", value))
            hive.commit(None)

        return True

    def update_uac(self, value):
        """Enable or disable the User Account Control by changing the value of
        the EnableLUA registry key

        value = 1 will enable the UAC
        value = 0 will disable the UAC

        Returns:
            True if the key is changed
            False if the key is unchanged
        """

        if value not in (0, 1):
            raise ValueError("Valid values for value parameter are 0 and 1")

        with self.open_hive('SOFTWARE', write=True) as hive:
            path = 'Microsoft/Windows/CurrentVersion/Policies/System'
            system = traverse(hive, path)

            enablelua = None
            for val in hive.node_values(system):
                if hive.value_key(val) == 'EnableLUA':
                    enablelua = val

            if enablelua is not None:
                if value == hive.value_dword(enablelua):
                    return False
            elif value == 1:
                return False

            hive.node_set_value(system, REG_DWORD('EnableLUA', value))
            hive.commit(None)

        return True

    def enum_users(self):
        """Returns:
            a map of RID->username for all users found on the system
            a list of RIDs of active users
            a list of RIDs of members of the Administrators group
        """

        users = {}
        active = []
        members = []

        # Under HKLM\SAM\SAM\Domains\Account\Users\%RID% there is an F field
        # that contains information about this user account. Bytes 56 & 57 are
        # the account type and status flags and the first bit in this flag is
        # the 'account disabled' bit:
        #
        # http://www.beginningtoseethelight.org/ntsecurity/index.htm
        #        #8603CF0AFBB170DD
        #
        disabled = lambda f: int(f[56].encode('hex'), 16) & 0x01

        def collect_group_members(hive, group, rid_node):
            """Enumerate group members"""
            c_val = hive.value_value(hive.node_get_value(rid_node, 'C'))[1]

            # Check http://pogostick.net/~pnh/ntpasswd/ for more info
            offset = struct.unpack('<I', c_val[0x28:0x2c])[0] + 0x34
            #size = struct.unpack('<I', c_val[0x2c:0x30])[0]
            count = struct.unpack('<I', c_val[0x30:0x34])[0]

            # Parse the sid array and get all members
            while len(members) < count:
                sections = struct.unpack('<B', c_val[offset+1:offset+2])[0]
                rid_offs = offset + 8 + (sections - 1) * 4
                rid = struct.unpack('<I', c_val[rid_offs:rid_offs+4])[0]
                members.append(rid)
                offset += sections * 4 + 8

        def collect_users(hive, username, rid_node):
            """Enumerate active users"""
            f_val = hive.value_value(hive.node_get_value(rid_node, 'F'))[1]
            rid = int(hive.node_name(rid_node), 16)

            if not disabled(f_val):
                active.append(rid)

            users[rid] = username

        self._foreach_account(
            userlist=[], useraction=collect_users,
            grouplist=[ADMINS], groupaction=collect_group_members)

        return (users, active, members)

    def reset_passwd(self, rid, v_field=None):
        r"""Reset the password for user 'user'. If v_field is not None, the
        value of key \HKEY_LOCAL_MACHINE\SAM\SAM\Domains\Account\Users\%RID%\V
        is replaced with this one, otherwise the LM and NT passwords of the
        original V field are cleared out. In any case, the value of the
        original V field is returned.
        """

        # This is a hack because python 2.x does not support assigning new
        # values to nonlocal variables
        parent = {}

        def update_v_field(hive, username, rid_node):
            """Updates the user's V field to reset the password"""
            assert 'old' not in parent, "Multiple users with same username"

            field = hive.node_get_value(rid_node, 'V')
            v_type, v_val = hive.value_value(field)
            assert v_type == 3L, "V field type (=%d) isn't REG_BINARY" % v_type

            new = v_field
            if new is None:
                # In order to reset the passwords, all we need to do is to zero
                # out the length fields for the LM password hash and the NT
                # password hash in the V field of the user's %RID% node. LM
                # password hash length field is at offset 0xa0 and NT password
                # hash length is at offset 0xac. See here for more info:
                #
                # http://www.beginningtoseethelight.org/ntsecurity/index.htm
                #        #D3BC3F5643A17823
                fmt = '%ds4x8s4x%ds' % (0xa0, len(v_val) - 0xb0)
                new = ("\x00" * 4).join(struct.unpack(fmt, v_val))

            hive.node_set_value(rid_node, REG_BINARY('V', new))
            hive.commit(None)
            parent['old'] = v_val

        self._foreach_account(True, userlist=[rid], useraction=update_v_field)

        assert 'old' in parent, "user whith RID: `%s' does not exist" % rid
        return parent['old']

    def reset_account(self, rid, activate=True):

        # This is a hack. I cannot assign a new value to nonlocal variable.
        # This is why I'm using a dict
        state = {}

        # Convert byte to int
        to_int = lambda b: int(b.encode('hex'), 16)

        # Under HKEY_LOCAL_MACHINE\SAM\SAM\Domains\Account\Users\%RID% there is
        # an F field that contains information about this user account. Bytes
        # 56 & 57 are the account type and status flags. The first bit is the
        # 'account disabled' bit:
        #
        # http://www.beginningtoseethelight.org/ntsecurity/index.htm
        #        #8603CF0AFBB170DD
        #
        isactive = lambda f: (to_int(f[56]) & 0x01) == 0

        def update_f_field(hive, username, rid_node):
            """Updates the user's F field to reset the account"""
            field = hive.node_get_value(rid_node, 'F')
            f_type, f_val = hive.value_value(field)
            assert f_type == 3L, "F field type (=%d) isn't REG_BINARY" % f_type

            state['old'] = isactive(f_val)
            if activate is state['old']:
                # nothing to do
                return

            mask = (lambda b: b & 0xfe) if activate else (lambda b: b | 0x01)
            new = struct.pack("56sB23s", f_val[:56], mask(to_int(f_val[56])),
                              f_val[57:])

            hive.node_set_value(rid_node, REG_BINARY('F', new))
            hive.commit(None)

        self._foreach_account(True, userlist=[rid], useraction=update_f_field)

        return state['old']

    def reset_first_logon_animation(self, activate=True):
        """Enable or disable the first-logon animation.

        The method return the old value
        """

        with self.open_hive('SOFTWARE', write=True) as hive:
            path = 'Microsoft/Windows/CurrentVersion/Policies/System'
            system = traverse(hive, path)
            try:
                val = hive.node_get_value(system, 'EnableFirstLogonAnimation')
                old = bool(hive.value_dword(val))

                # There is no need to reset the value
                if old is activate:
                    return old
            except RuntimeError:
                # The value is not present at all
                if activate is False:
                    return False
                old = False

            hive.node_set_value(system, REG_DWORD('EnableFirstLogonAnimation',
                                                  int(activate)))
            hive.commit(None)

        return old

    def _foreach_account(self, write=False, **kargs):
        """Performs an action on the RID node of a user or a group in the
        registry, for every user/group found in the userlist/grouplist.
        If userlist/grouplist is empty, it performs the action on all
        users/groups.
        The write flag determines if the registry is opened for reading or
        writing.
        """

        def parse_sam(ridlist, action, path):
            """Parse the registry users and groups nodes"""

            accounts = traverse(hive, 'SAM/Domains/' + path)
            names = hive.node_get_child(accounts, 'Names')

            # The RID (relative identifier) of each user/group is stored as the
            # type!!!! (not the value) of the default key of the node under
            # Names whose name is the username/groupname.

            for node in hive.node_children(names):
                name = hive.node_name(node)
                rid = hive.value_type(hive.node_get_value(node, ""))[0]

                if len(ridlist) != 0 and rid not in ridlist:
                    continue

                # if RID is 500 (=0x1f4), the corresponding node name is
                # '000001F4'
                key = ("%8.x" % rid).replace(' ', '0').upper()
                rid_node = hive.node_get_child(accounts, key)

                action(hive, name, rid_node)

        userlist = kargs['userlist'] if 'userlist' in kargs else None
        useraction = kargs['useraction'] if 'useraction' in kargs else None

        grouplist = kargs['grouplist'] if 'grouplist' in kargs else None
        groupaction = kargs['groupaction'] if 'groupaction' in kargs else None

        if userlist is not None:
            assert useraction is not None

        if grouplist is not None:
            assert groupaction is not None

        with self.open_hive('SAM', write) as hive:
            if userlist is not None:
                parse_sam(userlist, useraction, 'Account/Users')

            if grouplist is not None:
                parse_sam(grouplist, groupaction, 'Builtin/Aliases')
                parse_sam(grouplist, groupaction, 'Account/Aliases')

    def add_viostor(self):
        """Add the viostor driver to the critical device database and register
        the viostor service
        """

        path = r"system32\drivers\viostor.sys"
        pci = r"PCI\VEN_1AF4&DEV_1001&SUBSYS_00021AF4&REV_00\3&13c0b0c5&0&20"

        with self.open_hive('SYSTEM', write=True) as hive:

            # SYSTEM/CurrentControlSet/Control/CriticalDeviceDatabase
            control = traverse(hive, "/%s/Control" % self.current_control_set)
            cdd = safe_add_node(hive, control, 'CriticalDeviceDatabase')

            guid = "{4D36E97B-E325-11CE-BFC1-08002BE10318}"

            for subsys in '00000000', '00020000', '00021af4':
                name = "pci#ven_1af4&dev_1001&subsys_%s" % subsys

                node = safe_add_node(hive, cdd, name)

                hive.node_set_value(node, REG_SZ('ClassGUID', guid))
                hive.node_set_value(node, REG_SZ('Service', 'viostor'))

            # SYSTEM/CurrentContolSet/Services/viostor
            services = hive.root()
            for child in (self.current_control_set, 'Services'):
                services = hive.node_get_child(services, child)

            viostor = safe_add_node(hive, services, 'viostor')
            hive.node_set_value(viostor, REG_SZ('Group', 'SCSI miniport'))
            hive.node_set_value(viostor, REG_SZ('ImagePath', path))
            hive.node_set_value(viostor, REG_DWORD('ErrorControl', 1))
            hive.node_set_value(viostor, REG_DWORD('Start', 0))
            hive.node_set_value(viostor, REG_DWORD('Type', 1))
            hive.node_set_value(viostor, REG_DWORD('Tag', 0x21))

            params = safe_add_node(hive, viostor, 'Parameters')
            hive.node_set_value(params, REG_DWORD('BusType', 1))

            mts = safe_add_node(hive, params, 'MaxTransferSize')
            hive.node_set_value(mts,
                                REG_SZ('ParamDesc', 'Maximum Transfer Size'))
            hive.node_set_value(mts, REG_SZ('type', 'enum'))
            hive.node_set_value(mts, REG_SZ('default', '0'))

            enum = safe_add_node(hive, mts, 'enum')
            hive.node_set_value(enum, REG_SZ('0', '64  KB'))
            hive.node_set_value(enum, REG_SZ('1', '128 KB'))
            hive.node_set_value(enum, REG_SZ('2', '256 KB'))

            pnp_interface = safe_add_node(hive, params, 'PnpInterface')
            hive.node_set_value(pnp_interface, REG_DWORD('5', 1))
            enum = safe_add_node(hive, viostor, 'Enum')
            hive.node_set_value(enum, REG_SZ('0', pci))
            hive.node_set_value(enum, REG_DWORD('Count', 1))
            hive.node_set_value(enum, REG_DWORD('NextInstance', 1))

            hive.commit(None)

    def check_viostor_service(self):
        """Checks if the viostor service is installed"""

        service = '%s/Services/viostor' % self.current_control_set

        with self.open_hive('SYSTEM', write=False) as hive:
            return traverse(hive, service) is not None

    def update_devices_dirs(self, dirname, append=True):
        """Update the value of the DevicePath registry key. If the append flag
        is True, the dirname is appended to the list of devices directories,
        otherwise the value is overwritten.

        This function returns the old value of the registry key
        """

        with self.open_hive('SOFTWARE', write=True) as hive:

            current = traverse(hive, '/Microsoft/Windows/CurrentVersion')

            device_path = hive.node_get_value(current, 'DevicePath')
            regtype, value = hive.value_value(device_path)

            assert regtype == 2L, "Type (=%d) is not REG_EXPAND_SZ" % regtype

            # Remove the trailing '\x00' character
            old_value = value.decode('utf-16le')[:-1]

            new_value = "%s;%s" % (old_value, dirname) if append else dirname

            hive.node_set_value(current,
                                REG_EXPAND_SZ('DevicePath', new_value))
            hive.commit(None)

        return old_value

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :




__version__ = "0.7rc3"
__version_vcs_info__ = {
    'branch': 'release-0.7',
    'revid': 'd595255',
    'revno': 566}
__version_user_email__ = "skalkoto@grnet.gr"
__version_user_name__ = "Nikos Skalkotos"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module hosting the Disk class."""

from image_creator.util import get_command, try_fail_repeat, free_space, \
    FatalError, create_snapshot
from image_creator.bundle_volume import BundleVolume
from image_creator.image import Image

import stat
import os
import tempfile
import uuid
import shutil

dd = get_command('dd')
dmsetup = get_command('dmsetup')
losetup = get_command('losetup')
blockdev = get_command('blockdev')


def get_tmp_dir(default=None):
    """Check tmp directory candidates and return the one with the most
    available space.
    """
    if default is not None:
        return default

    TMP_CANDIDATES = ['/var/tmp', os.path.expanduser('~'), '/mnt']

    space = map(free_space, TMP_CANDIDATES)

    max_idx = 0
    max_val = space[0]
    for i, val in zip(range(len(space)), space):
        if val > max_val:
            max_val = val
            max_idx = i

    # Return the candidate path with more available space
    return TMP_CANDIDATES[max_idx]


class Disk(object):
    """This class represents a hard disk hosting an Operating System

    A Disk instance never alters the source media it is created from.
    Any change is done on a snapshot created by the device-mapper of
    the Linux kernel.
    """

    def __init__(self, source, output, tmp=None):
        """Create a new Disk instance out of a source media. The source
        media can be an image file, a block device or a directory.
        """
        self._cleanup_jobs = []
        self._images = []
        self._file = None
        self.source = source
        self.out = output
        self.meta = {}
        self.tmp = tempfile.mkdtemp(prefix='.snf_image_creator.',
                                    dir=get_tmp_dir(tmp))

        self._add_cleanup(shutil.rmtree, self.tmp)

    def _add_cleanup(self, job, *args):
        """Add a new job in the cleanup list"""
        self._cleanup_jobs.append((job, args))

    def _losetup(self, fname):
        """Setup a loop device and add it to the cleanup list. The loop device
        will be detached when cleanup is called.
        """
        loop = losetup('-f', '--show', fname)
        loop = loop.strip()  # remove the new-line char
        self._add_cleanup(try_fail_repeat, losetup, '-d', loop)
        return loop

    def _dir_to_disk(self):
        """Create a disk out of a directory"""
        if self.source == '/':
            bundle = BundleVolume(self.out, self.meta)
            image = '%s/%s.raw' % (self.tmp, uuid.uuid4().hex)

            def check_unlink(path):
                """Unlinks file if exists"""
                if os.path.exists(path):
                    os.unlink(path)

            self._add_cleanup(check_unlink, image)
            bundle.create_image(image)
            return image
        raise FatalError("Using a directory as media source is supported")

    def cleanup(self):
        """Cleanup internal data. This needs to be called before the
        program ends.
        """
        try:
            while len(self._images):
                image = self._images.pop()
                image.destroy()
        finally:
            # Make sure those are executed even if one of the device.destroy
            # methods throws exeptions.
            while len(self._cleanup_jobs):
                job, args = self._cleanup_jobs.pop()
                job(*args)

    @property
    def file(self):
        """Convert the source media into a file"""

        if self._file is not None:
            return self._file

        self.out.output("Examining source media `%s' ..." % self.source, False)
        mode = os.stat(self.source).st_mode
        if stat.S_ISDIR(mode):
            self.out.success('looks like a directory')
            self._file = self._dir_to_disk()
        elif stat.S_ISREG(mode):
            self.out.success('looks like an image file')
            self._file = self.source
        elif not stat.S_ISBLK(mode):
            raise FatalError("Invalid media source. Only block devices, "
                             "regular files and directories are supported.")
        else:
            self.out.success('looks like a block device')
            self._file = self.source

        return self._file

    def snapshot(self):
        """Creates a snapshot of the original source media of the Disk
        instance.
        """

        if self.source == '/':
            self.out.warn("Snapshotting ignored for host bundling mode.")
            return self.file

        # Examine media file
        mode = os.stat(self.file).st_mode

        self.out.output("Snapshotting media source ...", False)

        # Create a qcow2 snapshot for image files
        if not stat.S_ISBLK(mode):
            snapshot = create_snapshot(self.file, self.tmp)
            self._add_cleanup(os.unlink, snapshot)
            self.out.success('done')
            return snapshot

        # Create a device-mapper snapshot for block devices
        size = int(blockdev('--getsz', self.file))

        cowfd, cow = tempfile.mkstemp(dir=self.tmp)
        os.close(cowfd)
        self._add_cleanup(os.unlink, cow)
        # Create cow sparse file
        dd('if=/dev/null', 'of=%s' % cow, 'bs=512', 'seek=%d' % size)
        cowdev = self._losetup(cow)

        snapshot = 'snf-image-creator-snapshot-%s' % uuid.uuid4().hex
        tablefd, table = tempfile.mkstemp()
        try:
            try:
                os.write(tablefd, "0 %d snapshot %s %s n 8\n" %
                         (size, self.file, cowdev))
            finally:
                os.close(tablefd)

            dmsetup('create', snapshot, table)
            self._add_cleanup(try_fail_repeat, dmsetup, 'remove', snapshot)
        finally:
            os.unlink(table)
        self.out.success('done')
        return "/dev/mapper/%s" % snapshot

    def get_image(self, media, **kargs):
        """Returns a newly created Image instance."""

        image = Image(media, self.out, **kargs)
        self._images.append(image)
        image.enable()
        return image

    def destroy_image(self, image):
        """Destroys an Image instance previously created by get_image method.
        """
        self._images.remove(image)
        image.destroy()

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This modules provides the interface for working with the ./kamaki library.
The library is used to upload images to and register them with a Synnefo
deployment.
"""

import sys

from os.path import basename

from kamaki.cli.config import Config
from kamaki.clients import ClientError
from kamaki.clients.image import ImageClient
from kamaki.clients.pithos import PithosClient
from kamaki.clients.astakos import CachedAstakosClient as AstakosClient

try:
    config = Config()
except Exception as e:
    sys.stderr.write("Kamaki config error: %s\n" % str(e))
    sys.exit(1)


class Kamaki(object):
    """Wrapper class for the ./kamaki library"""
    CONTAINER = "images"

    @staticmethod
    def get_default_cloud_name():
        """Returns the name of the default cloud"""
        clouds = config.keys('cloud')
        default = config.get('global', 'default_cloud')
        if not default:
            return clouds[0] if len(clouds) else ""
        return default if default in clouds else ""

    @staticmethod
    def set_default_cloud(name):
        """Sets a cloud account as default"""
        config.set('global', 'default_cloud', name)
        config.write()

    @staticmethod
    def get_clouds():
        """Returns the list of available clouds"""
        names = config.keys('cloud')

        clouds = {}
        for name in names:
            clouds[name] = config.get('cloud', name)

        return clouds

    @staticmethod
    def get_cloud_by_name(name):
        """Returns a dictionary with cloud info"""
        return config.get('cloud', name)

    @staticmethod
    def save_cloud(name, url, token, description=""):
        """Save a new cloud account"""
        cloud = {'url': url, 'token': token}
        if len(description):
            cloud['description'] = description
        config.set('cloud', name, cloud)

        # Make the saved cloud the default one
        config.set('global', 'default_cloud', name)
        config.write()

    @staticmethod
    def remove_cloud(name):
        """Deletes an existing cloud from the ./Kamaki configuration file"""
        config.remove_option('cloud', name)
        config.write()

    @staticmethod
    def create_account(url, token):
        """Given a valid (URL, tokens) pair this method returns an Astakos
        client instance
        """
        client = AstakosClient(url, token)
        try:
            client.authenticate()
        except ClientError:
            return None

        return client

    @staticmethod
    def get_account(cloud_name):
        """Given a saved cloud name this method returns an Astakos client
        instance
        """
        cloud = config.get('cloud', cloud_name)
        assert cloud, "cloud: `%s' does not exist" % cloud_name
        assert 'url' in cloud, "url attr is missing in %s" % cloud_name
        assert 'token' in cloud, "token attr is missing in %s" % cloud_name

        return Kamaki.create_account(cloud['url'], cloud['token'])

    def __init__(self, account, output):
        """Create a Kamaki instance"""
        self.account = account
        self.out = output

        self.pithos = PithosClient(
            self.account.get_service_endpoints('object-store')['publicURL'],
            self.account.token,
            self.account.user_info()['id'],
            self.CONTAINER)

        self.image = ImageClient(
            self.account.get_service_endpoints('image')['publicURL'],
            self.account.token)

    def upload(self, file_obj, size=None, remote_path=None, hp=None, up=None):
        """Upload a file to Pithos+"""

        path = basename(file_obj.name) if remote_path is None else remote_path

        try:
            self.pithos.create_container(self.CONTAINER)
        except ClientError as e:
            if e.status != 202:  # Ignore container already exists errors
                raise e

        hash_cb = self.out.progress_generator(hp) if hp is not None else None
        upload_cb = self.out.progress_generator(up) if up is not None else None

        self.pithos.upload_object(path, file_obj, size, hash_cb, upload_cb)

        return "pithos://%s/%s/%s" % (self.account.user_info()['id'],
                                      self.CONTAINER, path)

    def register(self, name, location, metadata, public=False):
        """Register an image with Cyclades"""

        # Convert all metadata to strings
        str_metadata = {}
        for (key, value) in metadata.iteritems():
            str_metadata[str(key)] = str(value)
        is_public = 'true' if public else 'false'
        params = {'is_public': is_public, 'disk_format': 'diskdump'}
        return self.image.register(name, location, params, str_metadata)

    def share(self, location):
        """Share this file with all the users"""

        self.pithos.set_object_sharing(location, "*")

    def object_exists(self, location):
        """Check if an object exists in Pithos+"""

        try:
            self.pithos.get_object_info(location)
        except ClientError as e:
            if e.status == 404:  # Object not found error
                return False
            else:
                raise
        return True

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import image_creator

from setuptools import setup, find_packages

setup(
    name='snf_image_creator',
    version=image_creator.__version__,
    description='Command line tool for creating images',
    long_description=open('README.md').read(),
    url='https://code.grnet.gr/projects/snf-image-creator',
    author="Synnefo development team",
    author_email="synnefo-devel@googlegroups.com",
    license='BSD',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['sh', 'ansicolors', 'progress>=1.0.2'],
    entry_points={
        'console_scripts': [
                'snf-mkimage = image_creator.main:main',
                'snf-image-creator = image_creator.dialog_main:main']
    }
)
# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This package is intended to provide output classes for printing messages and
progress bars. The user can change the output behavior of the program by
subclassing the Output class and assigning the derived one as the output class
of the various parts of the image-creator package.
"""


class Output(object):
    """A class for printing program output"""
    def error(self, msg, new_line=True):
        """Print an error"""
        pass

    def warn(self, msg, new_line=True):
        """Print a warning"""
        pass

    def success(self, msg, new_line=True):
        """Print msg after an action is completed"""
        pass

    def output(self, msg='', new_line=True):
        """Print normal program output"""
        pass

    def cleanup(self):
        """Cleanup this output class"""
        pass

    def clear(self):
        """Clear the screen"""
        pass

    def _get_progress(self):
        """Returns a new Progress object"""
        progress = self._Progress
        progress.output = self
        return progress

    Progress = property(_get_progress)

    class _Progress(object):
        """Internal progress bar class"""
        def __init__(self, size, title, bar_type='default'):
            self.size = size
            self.bar_type = bar_type
            self.output.output("%s ..." % title, False)

        def goto(self, dest):
            """Move progress to a specific position"""
            pass

        def next(self):
            """Move progress a step forward"""
            pass

        def success(self, result):
            """Print a msg after an action is completed successfully"""
            self.output.success(result)

    def progress_generator(self, message):
        """A python generator for the progress bar class"""
        def generator(n):
            progressbar = self.Progress(n, message)

            for _ in range(n):
                yield
                progressbar.next()

            progressbar.success('done')
            yield
        return generator

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module implements the CompositeOutput output class"""

from image_creator.output import Output


class CompositeOutput(Output):
    """This class can be used to composite different outputs into a single one

    You may create an instance of this class and then add other output
    instances to it. Executing a method on this instance will cause the
    execution of the same method in each output instance that has been added to
    this one.
    """

    def __init__(self, outputs=[]):
        """Add initial output instances"""
        self._outputs = outputs

    def add(self, output):
        """Add another output instance"""
        self._outputs.append(output)

    def remove(self, output):
        """Remove an output instance"""
        self._outputs.remove(output)

    def error(self, msg, new_line=True):
        """Call the error method of each of the output instances"""
        for out in self._outputs:
            out.error(msg, new_line)

    def warn(self, msg, new_line=True):
        """Call the warn method of each of the output instances"""
        for out in self._outputs:
            out.warn(msg, new_line)

    def success(self, msg, new_line=True):
        """Call the success method of each of the output instances"""
        for out in self._outputs:
            out.success(msg, new_line)

    def output(self, msg='', new_line=True):
        """Call the output method of each of the output instances"""
        for out in self._outputs:
            out.output(msg, new_line)

    def cleanup(self):
        """Call the cleanup method of each of the output instances"""
        for out in self._outputs:
            out.cleanup()

    def clear(self):
        """Call the clear method of each of the output instances"""
        for out in self._outputs:
            out.clear()

    class _Progress(object):
        """Class used to composite different Progress objects"""

        def __init__(self, size, title, bar_type='default'):
            """Create a progress on each of the added output instances"""
            self._progresses = []
            for out in self.output._outputs:
                self._progresses.append(out.Progress(size, title, bar_type))

        def goto(self, dest):
            """Call the goto method of each of the progress instances"""
            for progress in self._progresses:
                progress.goto(dest)

        def next(self):
            """Call the next method of each of the progress instances"""
            for progress in self._progresses:
                progress.next()

        def success(self, result):
            """Call the success method of each of the progress instances"""
            for progress in self._progresses:
                progress.success(result)

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2014 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides an interface for the WinEXE utility"""

import subprocess
import time
import signal
from sh import which

from image_creator.util import FatalError


class WinexeTimeout(FatalError):
    """Raised when a WinExE command times-out"""
    pass


class WinEXE:
    """Wrapper class for the winexe command"""

    @staticmethod
    def is_installed(program='winexe'):
        return which(program) is not None

    def __init__(self, username, hostname, **kwargs):
        self._user = username
        self._host = hostname

        self._pass = kwargs['password'] if 'password' in kwargs else None
        self._prog = kwargs['progname'] if 'progname' in kwargs else 'winexe'

        # -U USERNAME[%PASSWORD]
        user = '%s%s' % (self._user, '%%%s' % self._pass if self._pass else "")
        self._opts = ['-U', user]

    def reset(self):
        """Reset all winexe options"""

        # -U USERNAME[%PASSWORD]
        user = '%s%s' % (self._user, '%%%s' % self._pass if self._pass else "")
        self._opts = ['-U', user]

    def runas(self, username, passwd=None):
        """Run command as this user"""

        opt = '--runas=%s%s' % (username,
                                '%%%s' % passwd if passwd is not None else "")
        self._opts.append(opt)
        return self

    def system(self):
        """Use SYSTEM account"""
        self._opts.append('--system')
        return self

    def no_pass(self):
        """Do not ask for password"""
        self._opts.append('--no-pass')
        return self

    def uninstall(self):
        """Uninstall winexe service after remote execution"""
        self._opts.append('--uninstall')
        return self

    def reinstall(self):
        """Reinstall winexe service before remote execution"""
        self._opts.append('--reinstall')
        return self

    def debug(self, level):
        """Set debug level"""
        self._opts.append('--debuglevel=%d' % level)
        return self

    def debug_stderr(self):
        """Send debug output to STDERR"""
        self._opts.append('--debug-stderr')

    def run(self, command, timeout=0):
        """Run a command on a remote windows system"""

        args = [self._prog] + self._opts + ["//%s" % self._host] + [command]
        run = subprocess.Popen(args, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)

        def handler(signum, frame):
            run.terminate()
            time.sleep(1)
            run.poll()
            if run.returncode is None:
                run.kill()
            run.wait()
            raise WinexeTimeout("Command: `%s' timed-out" % " ".join(args))

        signal.signal(signal.SIGALRM, handler)
        signal.alarm(timeout)
        stdout, stderr = run.communicate()
        rc = run.poll()
        signal.alarm(0)

        return (stdout, stderr, rc)

# vim: set sta sts=4 shiftwidth=4 sw=4 et ai :
